# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: deeplab2/model.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()

DESCRIPTOR = _descriptor.FileDescriptor(
    name='deeplab2/model.proto',
    package='deeplab2',
    syntax='proto2',
    serialized_options=b'P\001',
    serialized_pb=b'\n\x14\x64\x65\x65plab2/model.proto\x12\x08\x64\x65\x65plab2\"\xd7\x01\n\x0e\x44\x65\x63oderOptions\x12\x13\n\x0b\x66\x65\x61ture_key\x18\x01 \x01(\t\x12\x1d\n\x10\x64\x65\x63oder_channels\x18\x02 \x01(\x05:\x03\x32\x35\x36\x12\x33\n\x11\x64\x65\x63oder_conv_type\x18\x05 \x01(\t:\x18\x64\x65pthwise_separable_conv\x12\x1a\n\raspp_channels\x18\x03 \x01(\x05:\x03\x32\x35\x36\x12\x14\n\x0c\x61trous_rates\x18\x04 \x03(\x05\x12*\n\x1b\x61spp_use_only_1x1_proj_conv\x18\x06 \x01(\x08:\x05\x66\x61lse\"@\n\x0fLowLevelOptions\x12\x13\n\x0b\x66\x65\x61ture_key\x18\x01 \x01(\t\x12\x18\n\x10\x63hannels_project\x18\x02 \x01(\x05\"\xbd\x01\n\x0bHeadOptions\x12\x17\n\x0foutput_channels\x18\x01 \x01(\x05\x12\x15\n\rhead_channels\x18\x02 \x01(\x05\x12\x30\n\x0ehead_conv_type\x18\x03 \x01(\t:\x18\x64\x65pthwise_separable_conv\x12%\n\x1amax_value_after_activation\x18\x04 \x01(\x02:\x01\x30\x12%\n\x1amin_value_after_activation\x18\x05 \x01(\x02:\x01\x30\"\xac\x02\n\x0fInstanceOptions\x12\x14\n\x06\x65nable\x18\x01 \x01(\x08:\x04true\x12\x35\n\x12low_level_override\x18\x02 \x03(\x0b\x32\x19.deeplab2.LowLevelOptions\x12;\n\x19instance_decoder_override\x18\x03 \x01(\x0b\x32\x18.deeplab2.DecoderOptions\x12*\n\x0b\x63\x65nter_head\x18\x04 \x01(\x0b\x32\x15.deeplab2.HeadOptions\x12.\n\x0fregression_head\x18\x05 \x01(\x0b\x32\x15.deeplab2.HeadOptions\x12\x33\n\x14next_regression_head\x18\x06 \x01(\x0b\x32\x15.deeplab2.HeadOptions\"\x83\x0e\n\x0cModelOptions\x12\x38\n\x08\x62\x61\x63kbone\x18\x01 \x01(\x0b\x32&.deeplab2.ModelOptions.BackboneOptions\x12)\n\x07\x64\x65\x63oder\x18\x02 \x01(\x0b\x32\x18.deeplab2.DecoderOptions\x12=\n\ndeeplab_v3\x18\x03 \x01(\x0b\x32\'.deeplab2.ModelOptions.DeeplabV3OptionsH\x00\x12\x46\n\x0f\x64\x65\x65plab_v3_plus\x18\x04 \x01(\x0b\x32+.deeplab2.ModelOptions.DeeplabV3PlusOptionsH\x00\x12I\n\x10panoptic_deeplab\x18\x05 \x01(\x0b\x32-.deeplab2.ModelOptions.PanopticDeeplabOptionsH\x00\x12\x45\n\x0emotion_deeplab\x18\x07 \x01(\x0b\x32+.deeplab2.ModelOptions.MotionDeepLabOptionsH\x00\x12?\n\x0bmax_deeplab\x18\n \x01(\x0b\x32(.deeplab2.ModelOptions.MaXDeepLabOptionsH\x00\x12\x44\n\x0bvip_deeplab\x18\x0b \x01(\x0b\x32-.deeplab2.ModelOptions.PanopticDeeplabOptionsH\x00\x12\x1a\n\x12initial_checkpoint\x18\x06 \x01(\t\x12\x41\n3restore_semantic_last_layer_from_initial_checkpoint\x18\x08 \x01(\x08:\x04true\x12\x41\n3restore_instance_last_layer_from_initial_checkpoint\x18\t \x01(\x08:\x04true\x1a\xde\x02\n\x0f\x42\x61\x63kboneOptions\x12\x16\n\x04name\x18\x01 \x01(\t:\x08resnet50\x12\x19\n\routput_stride\x18\x02 \x01(\x05:\x02\x33\x32\x12\x1a\n\x12pretrained_weights\x18\x03 \x01(\t\x12%\n\x16use_squeeze_and_excite\x18\x04 \x01(\x08:\x05\x66\x61lse\x12\x1e\n\x13\x64rop_path_keep_prob\x18\x05 \x01(\x02:\x01\x31\x12$\n\x12\x64rop_path_schedule\x18\x06 \x01(\t:\x08\x63onstant\x12 \n\x15stem_width_multiplier\x18\x07 \x01(\x02:\x01\x31\x12$\n\x19\x62\x61\x63kbone_width_multiplier\x18\x08 \x01(\x02:\x01\x31\x12$\n\x19\x62\x61\x63kbone_layer_multiplier\x18\t \x01(\x02:\x01\x31\x12!\n\x15use_sac_beyond_stride\x18\n \x01(\x05:\x02-1\x1a\'\n\x10\x44\x65\x65plabV3Options\x12\x13\n\x0bnum_classes\x18\x01 \x01(\x05\x1aY\n\x14\x44\x65\x65plabV3PlusOptions\x12,\n\tlow_level\x18\x01 \x01(\x0b\x32\x19.deeplab2.LowLevelOptions\x12\x13\n\x0bnum_classes\x18\x02 \x01(\x05\x1a\xcc\x01\n\x16PanopticDeeplabOptions\x12,\n\tlow_level\x18\x01 \x03(\x0b\x32\x19.deeplab2.LowLevelOptions\x12+\n\x08instance\x18\x02 \x01(\x0b\x32\x19.deeplab2.InstanceOptions\x12,\n\rsemantic_head\x18\x03 \x01(\x0b\x32\x15.deeplab2.HeadOptions\x12)\n\ndepth_head\x18\x04 \x01(\x0b\x32\x15.deeplab2.HeadOptions\x1a\xcb\x01\n\x14MotionDeepLabOptions\x12,\n\tlow_level\x18\x01 \x03(\x0b\x32\x19.deeplab2.LowLevelOptions\x12+\n\x08instance\x18\x02 \x01(\x0b\x32\x19.deeplab2.InstanceOptions\x12,\n\rsemantic_head\x18\x03 \x01(\x0b\x32\x15.deeplab2.HeadOptions\x12*\n\x0bmotion_head\x18\x04 \x01(\x0b\x32\x15.deeplab2.HeadOptions\x1a\xb4\x01\n\x11MaXDeepLabOptions\x12/\n\x10pixel_space_head\x18\x01 \x01(\x0b\x32\x15.deeplab2.HeadOptions\x12\x36\n\x13\x61uxiliary_low_level\x18\x02 \x03(\x0b\x32\x19.deeplab2.LowLevelOptions\x12\x36\n\x17\x61uxiliary_semantic_head\x18\x03 \x01(\x0b\x32\x15.deeplab2.HeadOptionsB\x13\n\x11meta_architectureB\x02P\x01'
)

_DECODEROPTIONS = _descriptor.Descriptor(
    name='DecoderOptions',
    full_name='deeplab2.DecoderOptions',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='feature_key', full_name='deeplab2.DecoderOptions.feature_key', index=0,
            number=1, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='decoder_channels', full_name='deeplab2.DecoderOptions.decoder_channels', index=1,
            number=2, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=256,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='decoder_conv_type', full_name='deeplab2.DecoderOptions.decoder_conv_type', index=2,
            number=5, type=9, cpp_type=9, label=1,
            has_default_value=True, default_value=b"depthwise_separable_conv".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='aspp_channels', full_name='deeplab2.DecoderOptions.aspp_channels', index=3,
            number=3, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=256,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='atrous_rates', full_name='deeplab2.DecoderOptions.atrous_rates', index=4,
            number=4, type=5, cpp_type=1, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='aspp_use_only_1x1_proj_conv', full_name='deeplab2.DecoderOptions.aspp_use_only_1x1_proj_conv',
            index=5,
            number=6, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=False,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=35,
    serialized_end=250,
)

_LOWLEVELOPTIONS = _descriptor.Descriptor(
    name='LowLevelOptions',
    full_name='deeplab2.LowLevelOptions',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='feature_key', full_name='deeplab2.LowLevelOptions.feature_key', index=0,
            number=1, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='channels_project', full_name='deeplab2.LowLevelOptions.channels_project', index=1,
            number=2, type=5, cpp_type=1, label=1,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=252,
    serialized_end=316,
)

_HEADOPTIONS = _descriptor.Descriptor(
    name='HeadOptions',
    full_name='deeplab2.HeadOptions',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='output_channels', full_name='deeplab2.HeadOptions.output_channels', index=0,
            number=1, type=5, cpp_type=1, label=1,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='head_channels', full_name='deeplab2.HeadOptions.head_channels', index=1,
            number=2, type=5, cpp_type=1, label=1,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='head_conv_type', full_name='deeplab2.HeadOptions.head_conv_type', index=2,
            number=3, type=9, cpp_type=9, label=1,
            has_default_value=True, default_value=b"depthwise_separable_conv".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='max_value_after_activation', full_name='deeplab2.HeadOptions.max_value_after_activation', index=3,
            number=4, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(0),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='min_value_after_activation', full_name='deeplab2.HeadOptions.min_value_after_activation', index=4,
            number=5, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(0),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=319,
    serialized_end=508,
)

_INSTANCEOPTIONS = _descriptor.Descriptor(
    name='InstanceOptions',
    full_name='deeplab2.InstanceOptions',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='enable', full_name='deeplab2.InstanceOptions.enable', index=0,
            number=1, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=True,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='low_level_override', full_name='deeplab2.InstanceOptions.low_level_override', index=1,
            number=2, type=11, cpp_type=10, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='instance_decoder_override', full_name='deeplab2.InstanceOptions.instance_decoder_override', index=2,
            number=3, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='center_head', full_name='deeplab2.InstanceOptions.center_head', index=3,
            number=4, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='regression_head', full_name='deeplab2.InstanceOptions.regression_head', index=4,
            number=5, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='next_regression_head', full_name='deeplab2.InstanceOptions.next_regression_head', index=5,
            number=6, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=511,
    serialized_end=811,
)

_MODELOPTIONS_BACKBONEOPTIONS = _descriptor.Descriptor(
    name='BackboneOptions',
    full_name='deeplab2.ModelOptions.BackboneOptions',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='name', full_name='deeplab2.ModelOptions.BackboneOptions.name', index=0,
            number=1, type=9, cpp_type=9, label=1,
            has_default_value=True, default_value=b"resnet50".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='output_stride', full_name='deeplab2.ModelOptions.BackboneOptions.output_stride', index=1,
            number=2, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=32,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='pretrained_weights', full_name='deeplab2.ModelOptions.BackboneOptions.pretrained_weights', index=2,
            number=3, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='use_squeeze_and_excite', full_name='deeplab2.ModelOptions.BackboneOptions.use_squeeze_and_excite',
            index=3,
            number=4, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=False,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='drop_path_keep_prob', full_name='deeplab2.ModelOptions.BackboneOptions.drop_path_keep_prob', index=4,
            number=5, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(1),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='drop_path_schedule', full_name='deeplab2.ModelOptions.BackboneOptions.drop_path_schedule', index=5,
            number=6, type=9, cpp_type=9, label=1,
            has_default_value=True, default_value=b"constant".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='stem_width_multiplier', full_name='deeplab2.ModelOptions.BackboneOptions.stem_width_multiplier',
            index=6,
            number=7, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(1),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='backbone_width_multiplier',
            full_name='deeplab2.ModelOptions.BackboneOptions.backbone_width_multiplier', index=7,
            number=8, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(1),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='backbone_layer_multiplier',
            full_name='deeplab2.ModelOptions.BackboneOptions.backbone_layer_multiplier', index=8,
            number=9, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(1),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='use_sac_beyond_stride', full_name='deeplab2.ModelOptions.BackboneOptions.use_sac_beyond_stride',
            index=9,
            number=10, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=-1,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=1510,
    serialized_end=1860,
)

_MODELOPTIONS_DEEPLABV3OPTIONS = _descriptor.Descriptor(
    name='DeeplabV3Options',
    full_name='deeplab2.ModelOptions.DeeplabV3Options',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='num_classes', full_name='deeplab2.ModelOptions.DeeplabV3Options.num_classes', index=0,
            number=1, type=5, cpp_type=1, label=1,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=1862,
    serialized_end=1901,
)

_MODELOPTIONS_DEEPLABV3PLUSOPTIONS = _descriptor.Descriptor(
    name='DeeplabV3PlusOptions',
    full_name='deeplab2.ModelOptions.DeeplabV3PlusOptions',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='low_level', full_name='deeplab2.ModelOptions.DeeplabV3PlusOptions.low_level', index=0,
            number=1, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='num_classes', full_name='deeplab2.ModelOptions.DeeplabV3PlusOptions.num_classes', index=1,
            number=2, type=5, cpp_type=1, label=1,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=1903,
    serialized_end=1992,
)

_MODELOPTIONS_PANOPTICDEEPLABOPTIONS = _descriptor.Descriptor(
    name='PanopticDeeplabOptions',
    full_name='deeplab2.ModelOptions.PanopticDeeplabOptions',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='low_level', full_name='deeplab2.ModelOptions.PanopticDeeplabOptions.low_level', index=0,
            number=1, type=11, cpp_type=10, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='instance', full_name='deeplab2.ModelOptions.PanopticDeeplabOptions.instance', index=1,
            number=2, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='semantic_head', full_name='deeplab2.ModelOptions.PanopticDeeplabOptions.semantic_head', index=2,
            number=3, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='depth_head', full_name='deeplab2.ModelOptions.PanopticDeeplabOptions.depth_head', index=3,
            number=4, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=1995,
    serialized_end=2199,
)

_MODELOPTIONS_MOTIONDEEPLABOPTIONS = _descriptor.Descriptor(
    name='MotionDeepLabOptions',
    full_name='deeplab2.ModelOptions.MotionDeepLabOptions',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='low_level', full_name='deeplab2.ModelOptions.MotionDeepLabOptions.low_level', index=0,
            number=1, type=11, cpp_type=10, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='instance', full_name='deeplab2.ModelOptions.MotionDeepLabOptions.instance', index=1,
            number=2, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='semantic_head', full_name='deeplab2.ModelOptions.MotionDeepLabOptions.semantic_head', index=2,
            number=3, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='motion_head', full_name='deeplab2.ModelOptions.MotionDeepLabOptions.motion_head', index=3,
            number=4, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=2202,
    serialized_end=2405,
)

_MODELOPTIONS_MAXDEEPLABOPTIONS = _descriptor.Descriptor(
    name='MaXDeepLabOptions',
    full_name='deeplab2.ModelOptions.MaXDeepLabOptions',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='pixel_space_head', full_name='deeplab2.ModelOptions.MaXDeepLabOptions.pixel_space_head', index=0,
            number=1, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='auxiliary_low_level', full_name='deeplab2.ModelOptions.MaXDeepLabOptions.auxiliary_low_level',
            index=1,
            number=2, type=11, cpp_type=10, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='auxiliary_semantic_head', full_name='deeplab2.ModelOptions.MaXDeepLabOptions.auxiliary_semantic_head',
            index=2,
            number=3, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=2408,
    serialized_end=2588,
)

_MODELOPTIONS = _descriptor.Descriptor(
    name='ModelOptions',
    full_name='deeplab2.ModelOptions',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='backbone', full_name='deeplab2.ModelOptions.backbone', index=0,
            number=1, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='decoder', full_name='deeplab2.ModelOptions.decoder', index=1,
            number=2, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='deeplab_v3', full_name='deeplab2.ModelOptions.deeplab_v3', index=2,
            number=3, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='deeplab_v3_plus', full_name='deeplab2.ModelOptions.deeplab_v3_plus', index=3,
            number=4, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='panoptic_deeplab', full_name='deeplab2.ModelOptions.panoptic_deeplab', index=4,
            number=5, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='motion_deeplab', full_name='deeplab2.ModelOptions.motion_deeplab', index=5,
            number=7, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='max_deeplab', full_name='deeplab2.ModelOptions.max_deeplab', index=6,
            number=10, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='vip_deeplab', full_name='deeplab2.ModelOptions.vip_deeplab', index=7,
            number=11, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='initial_checkpoint', full_name='deeplab2.ModelOptions.initial_checkpoint', index=8,
            number=6, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='restore_semantic_last_layer_from_initial_checkpoint',
            full_name='deeplab2.ModelOptions.restore_semantic_last_layer_from_initial_checkpoint', index=9,
            number=8, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=True,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='restore_instance_last_layer_from_initial_checkpoint',
            full_name='deeplab2.ModelOptions.restore_instance_last_layer_from_initial_checkpoint', index=10,
            number=9, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=True,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
    ],
    extensions=[
    ],
    nested_types=[_MODELOPTIONS_BACKBONEOPTIONS, _MODELOPTIONS_DEEPLABV3OPTIONS, _MODELOPTIONS_DEEPLABV3PLUSOPTIONS,
                  _MODELOPTIONS_PANOPTICDEEPLABOPTIONS, _MODELOPTIONS_MOTIONDEEPLABOPTIONS,
                  _MODELOPTIONS_MAXDEEPLABOPTIONS, ],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name='meta_architecture', full_name='deeplab2.ModelOptions.meta_architecture',
            index=0, containing_type=None, fields=[]),
    ],
    serialized_start=814,
    serialized_end=2609,
)

_INSTANCEOPTIONS.fields_by_name['low_level_override'].message_type = _LOWLEVELOPTIONS
_INSTANCEOPTIONS.fields_by_name['instance_decoder_override'].message_type = _DECODEROPTIONS
_INSTANCEOPTIONS.fields_by_name['center_head'].message_type = _HEADOPTIONS
_INSTANCEOPTIONS.fields_by_name['regression_head'].message_type = _HEADOPTIONS
_INSTANCEOPTIONS.fields_by_name['next_regression_head'].message_type = _HEADOPTIONS
_MODELOPTIONS_BACKBONEOPTIONS.containing_type = _MODELOPTIONS
_MODELOPTIONS_DEEPLABV3OPTIONS.containing_type = _MODELOPTIONS
_MODELOPTIONS_DEEPLABV3PLUSOPTIONS.fields_by_name['low_level'].message_type = _LOWLEVELOPTIONS
_MODELOPTIONS_DEEPLABV3PLUSOPTIONS.containing_type = _MODELOPTIONS
_MODELOPTIONS_PANOPTICDEEPLABOPTIONS.fields_by_name['low_level'].message_type = _LOWLEVELOPTIONS
_MODELOPTIONS_PANOPTICDEEPLABOPTIONS.fields_by_name['instance'].message_type = _INSTANCEOPTIONS
_MODELOPTIONS_PANOPTICDEEPLABOPTIONS.fields_by_name['semantic_head'].message_type = _HEADOPTIONS
_MODELOPTIONS_PANOPTICDEEPLABOPTIONS.fields_by_name['depth_head'].message_type = _HEADOPTIONS
_MODELOPTIONS_PANOPTICDEEPLABOPTIONS.containing_type = _MODELOPTIONS
_MODELOPTIONS_MOTIONDEEPLABOPTIONS.fields_by_name['low_level'].message_type = _LOWLEVELOPTIONS
_MODELOPTIONS_MOTIONDEEPLABOPTIONS.fields_by_name['instance'].message_type = _INSTANCEOPTIONS
_MODELOPTIONS_MOTIONDEEPLABOPTIONS.fields_by_name['semantic_head'].message_type = _HEADOPTIONS
_MODELOPTIONS_MOTIONDEEPLABOPTIONS.fields_by_name['motion_head'].message_type = _HEADOPTIONS
_MODELOPTIONS_MOTIONDEEPLABOPTIONS.containing_type = _MODELOPTIONS
_MODELOPTIONS_MAXDEEPLABOPTIONS.fields_by_name['pixel_space_head'].message_type = _HEADOPTIONS
_MODELOPTIONS_MAXDEEPLABOPTIONS.fields_by_name['auxiliary_low_level'].message_type = _LOWLEVELOPTIONS
_MODELOPTIONS_MAXDEEPLABOPTIONS.fields_by_name['auxiliary_semantic_head'].message_type = _HEADOPTIONS
_MODELOPTIONS_MAXDEEPLABOPTIONS.containing_type = _MODELOPTIONS
_MODELOPTIONS.fields_by_name['backbone'].message_type = _MODELOPTIONS_BACKBONEOPTIONS
_MODELOPTIONS.fields_by_name['decoder'].message_type = _DECODEROPTIONS
_MODELOPTIONS.fields_by_name['deeplab_v3'].message_type = _MODELOPTIONS_DEEPLABV3OPTIONS
_MODELOPTIONS.fields_by_name['deeplab_v3_plus'].message_type = _MODELOPTIONS_DEEPLABV3PLUSOPTIONS
_MODELOPTIONS.fields_by_name['panoptic_deeplab'].message_type = _MODELOPTIONS_PANOPTICDEEPLABOPTIONS
_MODELOPTIONS.fields_by_name['motion_deeplab'].message_type = _MODELOPTIONS_MOTIONDEEPLABOPTIONS
_MODELOPTIONS.fields_by_name['max_deeplab'].message_type = _MODELOPTIONS_MAXDEEPLABOPTIONS
_MODELOPTIONS.fields_by_name['vip_deeplab'].message_type = _MODELOPTIONS_PANOPTICDEEPLABOPTIONS
_MODELOPTIONS.oneofs_by_name['meta_architecture'].fields.append(
    _MODELOPTIONS.fields_by_name['deeplab_v3'])
_MODELOPTIONS.fields_by_name['deeplab_v3'].containing_oneof = _MODELOPTIONS.oneofs_by_name['meta_architecture']
_MODELOPTIONS.oneofs_by_name['meta_architecture'].fields.append(
    _MODELOPTIONS.fields_by_name['deeplab_v3_plus'])
_MODELOPTIONS.fields_by_name['deeplab_v3_plus'].containing_oneof = _MODELOPTIONS.oneofs_by_name['meta_architecture']
_MODELOPTIONS.oneofs_by_name['meta_architecture'].fields.append(
    _MODELOPTIONS.fields_by_name['panoptic_deeplab'])
_MODELOPTIONS.fields_by_name['panoptic_deeplab'].containing_oneof = _MODELOPTIONS.oneofs_by_name['meta_architecture']
_MODELOPTIONS.oneofs_by_name['meta_architecture'].fields.append(
    _MODELOPTIONS.fields_by_name['motion_deeplab'])
_MODELOPTIONS.fields_by_name['motion_deeplab'].containing_oneof = _MODELOPTIONS.oneofs_by_name['meta_architecture']
_MODELOPTIONS.oneofs_by_name['meta_architecture'].fields.append(
    _MODELOPTIONS.fields_by_name['max_deeplab'])
_MODELOPTIONS.fields_by_name['max_deeplab'].containing_oneof = _MODELOPTIONS.oneofs_by_name['meta_architecture']
_MODELOPTIONS.oneofs_by_name['meta_architecture'].fields.append(
    _MODELOPTIONS.fields_by_name['vip_deeplab'])
_MODELOPTIONS.fields_by_name['vip_deeplab'].containing_oneof = _MODELOPTIONS.oneofs_by_name['meta_architecture']
DESCRIPTOR.message_types_by_name['DecoderOptions'] = _DECODEROPTIONS
DESCRIPTOR.message_types_by_name['LowLevelOptions'] = _LOWLEVELOPTIONS
DESCRIPTOR.message_types_by_name['HeadOptions'] = _HEADOPTIONS
DESCRIPTOR.message_types_by_name['InstanceOptions'] = _INSTANCEOPTIONS
DESCRIPTOR.message_types_by_name['ModelOptions'] = _MODELOPTIONS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DecoderOptions = _reflection.GeneratedProtocolMessageType('DecoderOptions', (_message.Message,), {
    'DESCRIPTOR': _DECODEROPTIONS,
    '__module__': 'deeplab2.model_pb2'
    # @@protoc_insertion_point(class_scope:deeplab2.DecoderOptions)
})
_sym_db.RegisterMessage(DecoderOptions)

LowLevelOptions = _reflection.GeneratedProtocolMessageType('LowLevelOptions', (_message.Message,), {
    'DESCRIPTOR': _LOWLEVELOPTIONS,
    '__module__': 'deeplab2.model_pb2'
    # @@protoc_insertion_point(class_scope:deeplab2.LowLevelOptions)
})
_sym_db.RegisterMessage(LowLevelOptions)

HeadOptions = _reflection.GeneratedProtocolMessageType('HeadOptions', (_message.Message,), {
    'DESCRIPTOR': _HEADOPTIONS,
    '__module__': 'deeplab2.model_pb2'
    # @@protoc_insertion_point(class_scope:deeplab2.HeadOptions)
})
_sym_db.RegisterMessage(HeadOptions)

InstanceOptions = _reflection.GeneratedProtocolMessageType('InstanceOptions', (_message.Message,), {
    'DESCRIPTOR': _INSTANCEOPTIONS,
    '__module__': 'deeplab2.model_pb2'
    # @@protoc_insertion_point(class_scope:deeplab2.InstanceOptions)
})
_sym_db.RegisterMessage(InstanceOptions)

ModelOptions = _reflection.GeneratedProtocolMessageType('ModelOptions', (_message.Message,), {

    'BackboneOptions': _reflection.GeneratedProtocolMessageType('BackboneOptions', (_message.Message,), {
        'DESCRIPTOR': _MODELOPTIONS_BACKBONEOPTIONS,
        '__module__': 'deeplab2.model_pb2'
        # @@protoc_insertion_point(class_scope:deeplab2.ModelOptions.BackboneOptions)
    })
    ,

    'DeeplabV3Options': _reflection.GeneratedProtocolMessageType('DeeplabV3Options', (_message.Message,), {
        'DESCRIPTOR': _MODELOPTIONS_DEEPLABV3OPTIONS,
        '__module__': 'deeplab2.model_pb2'
        # @@protoc_insertion_point(class_scope:deeplab2.ModelOptions.DeeplabV3Options)
    })
    ,

    'DeeplabV3PlusOptions': _reflection.GeneratedProtocolMessageType('DeeplabV3PlusOptions', (_message.Message,), {
        'DESCRIPTOR': _MODELOPTIONS_DEEPLABV3PLUSOPTIONS,
        '__module__': 'deeplab2.model_pb2'
        # @@protoc_insertion_point(class_scope:deeplab2.ModelOptions.DeeplabV3PlusOptions)
    })
    ,

    'PanopticDeeplabOptions': _reflection.GeneratedProtocolMessageType('PanopticDeeplabOptions', (_message.Message,), {
        'DESCRIPTOR': _MODELOPTIONS_PANOPTICDEEPLABOPTIONS,
        '__module__': 'deeplab2.model_pb2'
        # @@protoc_insertion_point(class_scope:deeplab2.ModelOptions.PanopticDeeplabOptions)
    })
    ,

    'MotionDeepLabOptions': _reflection.GeneratedProtocolMessageType('MotionDeepLabOptions', (_message.Message,), {
        'DESCRIPTOR': _MODELOPTIONS_MOTIONDEEPLABOPTIONS,
        '__module__': 'deeplab2.model_pb2'
        # @@protoc_insertion_point(class_scope:deeplab2.ModelOptions.MotionDeepLabOptions)
    })
    ,

    'MaXDeepLabOptions': _reflection.GeneratedProtocolMessageType('MaXDeepLabOptions', (_message.Message,), {
        'DESCRIPTOR': _MODELOPTIONS_MAXDEEPLABOPTIONS,
        '__module__': 'deeplab2.model_pb2'
        # @@protoc_insertion_point(class_scope:deeplab2.ModelOptions.MaXDeepLabOptions)
    })
    ,
    'DESCRIPTOR': _MODELOPTIONS,
    '__module__': 'deeplab2.model_pb2'
    # @@protoc_insertion_point(class_scope:deeplab2.ModelOptions)
})
_sym_db.RegisterMessage(ModelOptions)
_sym_db.RegisterMessage(ModelOptions.BackboneOptions)
_sym_db.RegisterMessage(ModelOptions.DeeplabV3Options)
_sym_db.RegisterMessage(ModelOptions.DeeplabV3PlusOptions)
_sym_db.RegisterMessage(ModelOptions.PanopticDeeplabOptions)
_sym_db.RegisterMessage(ModelOptions.MotionDeepLabOptions)
_sym_db.RegisterMessage(ModelOptions.MaXDeepLabOptions)

DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements Axial-ResNets proposed in Axial-DeepLab [1].

[1] Axial-Deeplab: Stand-Alone Axial-Attention for Panoptic Segmentation,
    ECCV 2020 Spotlight.
      Huiyu Wang, Yukun Zhu, Bradley Green, Hartwig Adam, Alan Yuille,
      Liang-Chieh Chen.
"""

import tensorflow as tf

from deeplab2.model import utils
from deeplab2.model.layers import activations
from deeplab2.model.layers import axial_block_groups
from deeplab2.model.layers import convolutions
from deeplab2.model.layers import resized_fuse
from deeplab2.model.layers import stems

# Add a suffix in layer names that indicate if the current layer is a part of
# the backbone or an extra layer, i.e. if the current layer will be pretrained
# or not. This name will be used when we apply 10x larger learning rates for
# extra parameters that have not been pretrained, in panoptic segmentation.
# This keyword is reserved and should not be a part of the variable names in a
# classification pretrained backbone.
EXTRA = 'extra'
# Similarly, we will apply 10x larger learning rates on the memory feature.
# This global variable name will be accessed when we build the optimizers. This
# keyword is reserved and should not be a part of the variable names in a
# classification pretrained backbone.
MEMORY_FEATURE = 'memory_feature'


class AxialResNet(tf.keras.Model):
    """An Axial-ResNet model as proposed in Axial-DeepLab [1] and MaX-DeepLab [2].

    An Axial-ResNet [1] replaces 3x3 convolutions in a Resnet by axial-attention
    layers. A dual-path transformer [2] and a stacked decoder [2] can be used
    optionally. In addition, this class supports scaling models with SWideRNet [3]
    and augmenting convolutions with Switchable Atrous Convolution [4].

    Reference:
    [1] Axial-Deeplab: Stand-Alone Axial-Attention for Panoptic Segmentation,
        ECCV 2020 Spotlight. https://arxiv.org/abs/2003.07853
          Huiyu Wang, Yukun Zhu, Bradley Green, Hartwig Adam, Alan Yuille,
          Liang-Chieh Chen.
    [2] MaX-DeepLab: "End-to-End Panoptic Segmentation with Mask Transformers",
        CVPR 2021. https://arxiv.org/abs/2012.00759
          Huiyu Wang, Yukun Zhu, Hartwig Adam, Alan Yuille, Liang-Chieh Chen.
    [3] Scaling Wide Residual Networks for Panoptic Segmentation,
        https://arxiv.org/abs/2011.11675
          Liang-Chieh Chen, Huiyu Wang, Siyuan Qiao.
    [4] DetectoRS: Detecting Objects with Recursive Feature Pyramid and Switchable
        Atrous Convolution, CVPR 2021. https://arxiv.org/abs/2006.02334
          Siyuan Qiao, Liang-Chieh Chen, Alan Yuille.
    """

    def __init__(self,
                 name,
                 num_blocks=(3, 4, 6, 3),
                 backbone_layer_multiplier=1.0,
                 width_multiplier=1.0,
                 stem_width_multiplier=1.0,
                 output_stride=16,
                 classification_mode=False,
                 backbone_type='resnet_beta',
                 use_axial_beyond_stride=16,
                 backbone_use_transformer_beyond_stride=32,
                 extra_decoder_use_transformer_beyond_stride=32,
                 backbone_decoder_num_stacks=0,
                 backbone_decoder_blocks_per_stage=1,
                 extra_decoder_num_stacks=0,
                 extra_decoder_blocks_per_stage=1,
                 max_num_mask_slots=128,
                 num_mask_slots=128,
                 memory_channels=256,
                 base_transformer_expansion=1.0,
                 global_feed_forward_network_channels=256,
                 high_resolution_output_stride=4,
                 activation='relu',
                 block_group_config=None,
                 bn_layer=tf.keras.layers.BatchNormalization,
                 conv_kernel_weight_decay=0.0):
        """Initializes an AxialResNet model.

        Args:
          name: A string, the name of the model.
          num_blocks: A list of 4 integers. It denotes the number of blocks to
            include in the last 4 stages or block groups. Each group consists of
            blocks that output features of the same resolution. Defaults to (3, 4,
            6, 3) as in MaX-DeepLab-S.
          backbone_layer_multiplier: A float, layer_multiplier for the backbone,
            excluding the STEM. This flag controls the number of layers. Defaults to
            1.0 as in MaX-DeepLab-S.
          width_multiplier: A float, the channel multiplier for the block groups.
            Defaults to 1.0 as in MaX-DeepLab-S.
          stem_width_multiplier: A float, the channel multiplier for stem
            convolutions. Defaults to 1.0 as in MaX-DeepLab-S.
          output_stride: An integer, the maximum ratio of input to output spatial
            resolution. Defaults to 16 as in MaX-DeepLab-S.
          classification_mode: A boolean, whether to perform in a classification
            mode. If it is True, this function directly returns backbone feature
            endpoints. Note that these feature endpoints can also be used directly
            for Panoptic-DeepLab or Motion-DeepLab. If it is False, this function
            builds MaX-DeepLab extra decoder layers and extra transformer layers.
            Defaults to False as in MaX-DeepLab.
          backbone_type: A string, the type of backbone. Supports 'resnet',
            'resnet_beta', and 'wider_resnet'. It controls both the stem type and
            the residual block type. Defaults to 'resnet_beta' as in MaX-DeepLab-S.
          use_axial_beyond_stride: An integer, the stride beyond which we use axial
            attention. Set to 0 if no axial attention is desired. Defaults to 16 as
            in MaX-DeepLab.
          backbone_use_transformer_beyond_stride: An integer, the stride beyond
            which we use a memory path transformer block on top of a regular pixel
            path block, in the backbone. Set to 0 if no transformer block is desired
            in the backbone. Defaults to 32 as in MaX-DeepLab-S.
          extra_decoder_use_transformer_beyond_stride: An integer, the stride beyond
            which we use a memory path transformer block on top of a regular pixel
            path block, in the extra decoder stages. Set to 0 if no transformer
            block is desired in the extra decoder stages. Defaults to 32 as in
            MaX-DeepLab-S.
          backbone_decoder_num_stacks: An integer, the number of decoder stacks
            (introduced in MaX-DeepLab) that we use in the backbone. The stacked
            decoders are applied in a stacked hour-glass style. Defaults to 0 as in
            MaX-DeepLab-S.
          backbone_decoder_blocks_per_stage: An integer, the number of consecutive
            residual blocks to apply for each decoder stage, in the backbone.
            Defaults to 1 as in MaX-DeepLab-S.
          extra_decoder_num_stacks: An integer, the number of decoder stacks
            (introduced in MaX-DeepLab) that we use in the extra decoder layers. It
            is different from backbone_decoder_blocks_per_stage in that the extra
            decoder stacks will be trained from scratch on segmentation tasks,
            instead of pretrained on ImageNet classification. Defaults to 0 as in
            MaX-DeepLab-S.
          extra_decoder_blocks_per_stage: An integer, the number of consecutive
            residual blocks to apply for each decoder stage, in the extra decoder
            stages. Defaults to 1 as in MaX-DeepLab-S.
          max_num_mask_slots: An integer, the maximum possible number of mask slots
            that will be used. This will be used in a pretraining-finetuning use
            case with different num_mask_slots: We can set max_num_mask_slots to the
            maximum possible num_mask_slots, and then the saved checkpoint can be
            loaded for finetuning with a different num_mask_slots. Defaults to 128
            as in MaX-DeepLab.
          num_mask_slots: An integer, the number of mask slots that will be used.
            Defaults to 128 as in MaX-DeepLab-S.
          memory_channels: An integer, the number of channels for the whole memory
            path. Defaults to 256 as in MaX-DeepLab-S.
          base_transformer_expansion: A float, the base width expansion rate for
            transformer layers. Defaults to 1.0 as in MaX-DeepLab-S.
          global_feed_forward_network_channels: An integer, the number of channels
            in the final global feed forward network, i.e. the mask feature head and
            the mask class head. Defaults to 256 as in MaX-DeepLab-S.
          high_resolution_output_stride: An integer, the final decoding output
            stride. Defaults to 4 as in MaX-DeepLab-S.
          activation: A string, type of activation function to apply. Support
            'relu', 'swish' (or 'silu'), 'gelu', 'approximated_gelu', and 'elu'.
          block_group_config: An argument dictionary that will be passed to
            block_group.
          bn_layer: An optional tf.keras.layers.Layer that computes the
            normalization (default: tf.keras.layers.BatchNormalization).
          conv_kernel_weight_decay: A float, the weight decay for convolution
            kernels.

        Raises:
          ValueError: If backbone_type is not one of 'resnet', 'resnet_beta', or
            'wider_resnet'.
          ValueError: If extra_decoder_blocks_per_stage is not greater than zero.
        """
        super(AxialResNet, self).__init__(name=name)

        if extra_decoder_blocks_per_stage <= 0:
            raise ValueError(
                'Extra_decoder_blocks_per_stage should be great than zero.')
        if block_group_config is None:
            block_group_config = {}

        # Compute parameter lists for block_groups. We consider five stages so that
        # it is general enough to cover fully axial resnets and wider resnets.
        total_strides_list = [1, 2, 4, 8, 16]

        # Append 3 blocks for the first stage of fully axial resnets and wider
        # resnets.
        num_blocks_list = [3] + utils.scale_int_list(list(num_blocks),
                                                     backbone_layer_multiplier)
        strides_list = [2] * 5

        # Expand the transformer and the block filters with the stride.
        transformer_expansions_list = []
        filters_list = []
        for index, stride in enumerate(total_strides_list):
            # Reduce the number of channels when we apply transformer to low level
            # features (stride = 2, 4, or 8). The base_transformer_expansion is used
            # for stride = 16, i.e. the standard output_stride for MaX-DeepLab-S.
            transformer_expansions_list.append(base_transformer_expansion * stride /
                                               16.0)
            # Compute the base number of filters in each stage. For example, the last
            # stage of ResNet50 has an input stride of 16, then we compute the base
            # number of filters for a bottleneck block as 16 * 32 = 512, which is the
            # number of filters for the 3x3 convolution in those blocks.
            if backbone_type == 'wider_resnet' and index == 0:
                # SWideRNet variants use stem_width_multiplier for the first block.
                filters_list.append(int(round(stride * 32 * stem_width_multiplier)))
            else:
                filters_list.append(int(round(stride * 32 * width_multiplier)))

        self._num_mask_slots = None
        # Initialize memory_feature only when a transformer block is used.
        self._use_memory_feature = (backbone_use_transformer_beyond_stride or
                                    (extra_decoder_use_transformer_beyond_stride and
                                     (not classification_mode)))
        if self._use_memory_feature:
            self._memory_feature_shape = (1, max_num_mask_slots, memory_channels)
            self._memory_feature_initializer = (
                tf.keras.initializers.TruncatedNormal(stddev=1.0))
            self._memory_feature_regularizer = tf.keras.regularizers.l2(
                conv_kernel_weight_decay)
            if num_mask_slots:
                self._num_mask_slots = num_mask_slots

        # Use a convolutional stem except fully axial cases.
        stem_channels = int(round(64 * stem_width_multiplier))
        self._activation_fn = activations.get_activation(activation)
        if use_axial_beyond_stride == 1:
            self._stem = tf.identity
            first_block_index = 0
        elif backbone_type.lower() == 'wider_resnet':
            self._stem = convolutions.Conv2DSame(
                output_channels=stem_channels,
                kernel_size=3,
                name='stem',
                strides=2,
                use_bias=False,
                use_bn=True,
                bn_layer=bn_layer,
                activation='none',
                conv_kernel_weight_decay=conv_kernel_weight_decay)
            # Wider ResNet has five residual block stages, so we start from index 0.
            first_block_index = 0
            # Since we have applied the first strided convolution here, we do not use
            # a stride for the first stage (which will operate on stride 2).
            strides_list[0] = 1
            total_strides_list[0] = 2
        elif backbone_type.lower() == 'resnet_beta':
            self._stem = stems.InceptionSTEM(
                bn_layer=bn_layer,
                width_multiplier=stem_width_multiplier,
                conv_kernel_weight_decay=conv_kernel_weight_decay,
                activation=activation)
            first_block_index = 1
        elif backbone_type.lower() == 'resnet':
            self._stem = convolutions.Conv2DSame(
                output_channels=stem_channels,
                kernel_size=7,
                name='stem',
                strides=2,
                use_bias=False,
                use_bn=True,
                bn_layer=bn_layer,
                activation='none',
                conv_kernel_weight_decay=conv_kernel_weight_decay)
            first_block_index = 1
        else:
            raise ValueError(backbone_type + ' is not supported.')

        self._first_block_index = first_block_index
        # Apply standard ResNet block groups. We use first_block_index to
        # distinguish models with 4 stages and those with 5 stages.
        for index in range(first_block_index, 5):
            current_name = '_stage{}'.format(index + 1)
            utils.safe_setattr(self, current_name, axial_block_groups.BlockGroup(
                filters=filters_list[index],
                num_blocks=num_blocks_list[index],
                name=utils.get_layer_name(current_name),
                original_resnet_stride=strides_list[index],
                original_resnet_input_stride=total_strides_list[index],
                output_stride=output_stride,
                backbone_type=backbone_type,
                use_axial_beyond_stride=use_axial_beyond_stride,
                use_transformer_beyond_stride=(
                    backbone_use_transformer_beyond_stride),
                transformer_expansion=transformer_expansions_list[index],
                activation=activation,
                bn_layer=bn_layer,
                conv_kernel_weight_decay=conv_kernel_weight_decay,
                **block_group_config))
        self._backbone_decoder_num_stacks = backbone_decoder_num_stacks
        self._classification_mode = classification_mode
        self._extra_decoder_num_stacks = extra_decoder_num_stacks
        self._output_stride = output_stride
        self._high_resolution_output_stride = high_resolution_output_stride
        self._width_multiplier = width_multiplier
        self._activation = activation
        self._bn_layer = bn_layer
        self._conv_kernel_weight_decay = conv_kernel_weight_decay
        self._backbone_use_transformer_beyond_stride = (
            backbone_use_transformer_beyond_stride)
        self._extra_decoder_use_transformer_beyond_stride = (
            extra_decoder_use_transformer_beyond_stride)

        # Keep track of the current stack so that we know when to stop.
        current_stack = 0
        # Track whether we are building the backbone. This will affect the backbone
        # related arguments, local learning rate, and so on.
        current_is_backbone = True

        if backbone_decoder_num_stacks == 0:
            # No stacked decoder is used in the backbone, so we have finished building
            # the backbone. We either return the classification endpoints, or continue
            # building a non-backbone decoder for panoptic segmentation.
            if self._classification_mode:
                return
            else:
                current_is_backbone = False
        if not current_is_backbone:
            # Now that we have finished building the backbone and no stacked decoder
            # is used in the backbone, so we start to build extra (i.e., non-backbone)
            # layers for panoptic segmentation.
            current_name = '_stage5_' + EXTRA
            utils.safe_setattr(
                self, current_name, axial_block_groups.BlockGroup(
                    filters=filters_list[-1],
                    num_blocks=extra_decoder_blocks_per_stage,
                    name=utils.get_layer_name(current_name),
                    original_resnet_stride=1,
                    original_resnet_input_stride=32,
                    output_stride=output_stride,
                    backbone_type=backbone_type,
                    use_axial_beyond_stride=use_axial_beyond_stride,
                    use_transformer_beyond_stride=(
                        extra_decoder_use_transformer_beyond_stride),
                    transformer_expansion=base_transformer_expansion,
                    activation=activation,
                    bn_layer=bn_layer,
                    conv_kernel_weight_decay=conv_kernel_weight_decay,
                    **block_group_config))

        # Compute parameter lists for stacked decoder.
        total_decoder_num_stacks = (
                backbone_decoder_num_stacks + extra_decoder_num_stacks)

        # Use a function to compute the next stride.
        # next_stride_fn = lambda x: x // 2
        def next_stride_fn(x):
            return x // 2

        current_decoder_stride = output_stride
        decoder_stage = 0

        # Exit if we have enough stacks and reach the decoding output stride.
        while (current_stack < total_decoder_num_stacks or
               current_decoder_stride > high_resolution_output_stride):
            decoder_stage += 1
            current_decoder_stride = next_stride_fn(current_decoder_stride)

            if current_decoder_stride == output_stride:
                current_stack += 1
                # Always use blocks from the last resnet stage if the current stride is
                # output stride (the largest stride).
                original_resnet_input_stride = 32

                # Switch the decoder direction if we reach the largest stride.
                def next_stride_fn(x):
                    return x // 2
            else:
                original_resnet_input_stride = current_decoder_stride

            # Scale channels according to the strides.
            decoder_channels = original_resnet_input_stride * 64 * width_multiplier
            current_transformer_expansion = (
                    base_transformer_expansion * current_decoder_stride / 16.0)

            # Apply a decoder block group for building the backbone.
            if current_is_backbone:
                current_name = '_decoder_stage{}'.format(decoder_stage)
                utils.safe_setattr(
                    self, current_name, axial_block_groups.BlockGroup(
                        filters=decoder_channels // 4,
                        num_blocks=backbone_decoder_blocks_per_stage,
                        name=utils.get_layer_name(current_name),
                        original_resnet_stride=1,
                        original_resnet_input_stride=original_resnet_input_stride,
                        output_stride=output_stride,
                        backbone_type=backbone_type,
                        use_axial_beyond_stride=use_axial_beyond_stride,
                        use_transformer_beyond_stride=(
                            backbone_use_transformer_beyond_stride),
                        transformer_expansion=current_transformer_expansion,
                        activation=activation,
                        bn_layer=bn_layer,
                        conv_kernel_weight_decay=conv_kernel_weight_decay,
                        **block_group_config))

            if (current_decoder_stride == output_stride and
                    current_stack == backbone_decoder_num_stacks):
                # Now that we have finished building the backbone, we either return the
                # classification endpoints, or continue building a non-backbone decoder
                # for panoptic segmentation.
                if classification_mode:
                    return
                else:
                    current_is_backbone = False

            # Apply a decoder block group for building the extra layers.
            if not current_is_backbone:
                # Continue building an extra (i.e., non-backbone) decoder for panoptic
                # segmentation.
                current_name = '_decoder_stage{}_{}'.format(decoder_stage, EXTRA)
                utils.safe_setattr(
                    self, current_name, axial_block_groups.BlockGroup(
                        filters=decoder_channels // 4,
                        num_blocks=extra_decoder_blocks_per_stage,
                        name=utils.get_layer_name(current_name),
                        original_resnet_stride=1,
                        original_resnet_input_stride=original_resnet_input_stride,
                        output_stride=output_stride,
                        backbone_type=backbone_type,
                        use_axial_beyond_stride=use_axial_beyond_stride,
                        use_transformer_beyond_stride=(
                            extra_decoder_use_transformer_beyond_stride),
                        transformer_expansion=current_transformer_expansion,
                        activation=activation,
                        bn_layer=bn_layer,
                        conv_kernel_weight_decay=conv_kernel_weight_decay,
                        **block_group_config))
            if current_decoder_stride == high_resolution_output_stride:
                def next_stride_fn(x):
                    return x * 2

        # Assert that we have already returned if we are building a classifier.
        assert not classification_mode
        if (backbone_use_transformer_beyond_stride or
                extra_decoder_use_transformer_beyond_stride):
            # Build extra memory path feed forward networks for the class feature and
            # the mask feature.
            current_name = '_class_feature_' + EXTRA
            utils.safe_setattr(
                self, current_name, convolutions.Conv1D(
                    global_feed_forward_network_channels,
                    utils.get_layer_name(current_name),
                    use_bias=False,
                    use_bn=True,
                    bn_layer=bn_layer,
                    activation=activation,
                    conv_kernel_weight_decay=conv_kernel_weight_decay))
            current_name = '_mask_feature_' + EXTRA
            utils.safe_setattr(
                self, current_name, convolutions.Conv1D(
                    global_feed_forward_network_channels,
                    utils.get_layer_name(current_name),
                    use_bias=False,
                    use_bn=True,
                    bn_layer=bn_layer,
                    activation=activation,
                    conv_kernel_weight_decay=conv_kernel_weight_decay))

    def build(self, input_shape):
        """Builds model weights and input shape dependent sub-layers."""
        if self._use_memory_feature:
            self._memory_feature = self.add_weight(
                name=MEMORY_FEATURE,
                shape=self._memory_feature_shape,
                initializer=self._memory_feature_initializer,
                regularizer=self._memory_feature_regularizer)
        else:
            self._memory_feature = None

        # Go through the loop to build the ResizedFuse layers.
        current_stack = 0
        # Track whether we are building the backbone. This will affect the backbone
        # related arguments, local learning rate, and so on.
        current_is_backbone = self._backbone_decoder_num_stacks != 0
        total_decoder_num_stacks = (
                self._backbone_decoder_num_stacks + self._extra_decoder_num_stacks)

        def next_stride_fn(x):
            return x // 2
        current_decoder_stride = self._output_stride
        decoder_stage = 0
        while (current_stack < total_decoder_num_stacks or
               current_decoder_stride > self._high_resolution_output_stride):
            decoder_stage += 1
            current_decoder_stride = next_stride_fn(current_decoder_stride)
            if current_decoder_stride == self._output_stride:
                current_stack += 1
                original_resnet_input_stride = 32

                def next_stride_fn(x):
                    return x // 2
            else:
                original_resnet_input_stride = current_decoder_stride
            # Compute the decoder_channels according to original_resnet_input_stride.
            # For example, at stride 4 with width multiplier = 1, we use 4 * 64 = 256
            # channels, which is the same as a standard ResNet.
            decoder_channels = int(round(
                original_resnet_input_stride * 64 * self._width_multiplier))
            decoder_height, decoder_width = utils.scale_mutable_sequence(
                input_shape[1:3], 1.0 / current_decoder_stride)
            if current_is_backbone:
                current_name = '_decoder_stage{}_resized_fuse'.format(decoder_stage)
            else:
                current_name = '_decoder_stage{}_{}_resized_fuse'.format(
                    decoder_stage, EXTRA)
            utils.safe_setattr(
                self, current_name, resized_fuse.ResizedFuse(
                    name=utils.get_layer_name(current_name),
                    height=decoder_height,
                    width=decoder_width,
                    num_channels=decoder_channels,
                    activation=self._activation,
                    bn_layer=self._bn_layer,
                    conv_kernel_weight_decay=self._conv_kernel_weight_decay))
            if (current_decoder_stride == self._output_stride and
                    current_stack == self._backbone_decoder_num_stacks):
                # Now that we have finished building the backbone, we either return the
                # classification endpoints, or continue building a non-backbone decoder
                # for panoptic segmentation.
                if self._classification_mode:
                    return
                current_is_backbone = False
            if current_decoder_stride == self._high_resolution_output_stride:
                def next_stride_fn(x):
                    return x * 2

    def call_encoder_before_stacked_decoder(self, inputs, training=False):
        """Performs a forward pass of the encoder before stacking decoders.

        Args:
          inputs: An input [batch, height, width, channel] tensor.
          training: A boolean, whether the model is in training mode.

        Returns:
          current_output: An output tensor with shape [batch, new_height, new_width,
            new_channel].
          activated_output: An activated output tensor with shape [batch,
            new_height, new_width, new_channel].
          memory_feature: None if no transformer is used. A [batch, num_memory,
            memory_channel] tensor if transformer is used.
          endpoints: A dict, the network endpoints that might be used by DeepLab.
        """
        memory_feature = self._memory_feature
        if self._use_memory_feature:
            if self._num_mask_slots:
                memory_feature = self._memory_feature[:, :self._num_mask_slots, :]
            memory_feature = tf.tile(memory_feature,
                                     [tf.shape(inputs)[0], 1, 1])

        endpoints = {}
        current_output = self._stem(inputs)
        activated_output = self._activation_fn(current_output)
        endpoints['stage1'] = current_output
        endpoints['res1'] = activated_output

        # Apply standard ResNet block groups. We use first_block_index to
        # distinguish models with 4 stages and those with 5 stages.
        for index in range(self._first_block_index, 5):
            current_name = '_stage{}'.format(index + 1)
            current_output, memory_feature = (
                getattr(self, current_name)(
                    (current_output, memory_feature), training=training))
            activated_output = self._activation_fn(current_output)
            endpoints[utils.get_layer_name(current_name)] = current_output
            activated_output_name = 'res{}'.format(index + 1)
            endpoints[activated_output_name] = activated_output
        return current_output, activated_output, memory_feature, endpoints

    def call_stacked_decoder(self,
                             current_output,
                             activated_output,
                             memory_feature,
                             endpoints,
                             training=False):
        """Performs a forward pass of the stacked decoders.

        Args:
          current_output: An output tensor with shape [batch, new_height, new_width,
            new_channel].
          activated_output: An activated output tensor with shape [batch,
            new_height, new_width, new_channel].
          memory_feature: None if no transformer is used. A [batch, num_memory,
            memory_channel] tensor if transformer is used.
          endpoints: A dict, the network endpoints that might be used by DeepLab.
          training: A boolean, whether the model is in training mode.

        Returns:
          memory_feature: None if no transformer is used. A [batch, num_memory,
            memory_channel] tensor if transformer is used.
          high_resolution_outputs: A list of decoded tensors with
            high_resolution_output_stride.
          backbone_output: An output tensor of the backbone, with output_stride.
          endpoints: A dict, the network endpoints that might be used by DeepLab.
        """
        # Keep track of the current stack so that we know when to stop.
        current_stack = 0
        # Track whether we are building the backbone. This will affect the backbone
        # related arguments, local learning rate, and so on.
        current_is_backbone = True
        high_resolution_outputs = []

        if self._backbone_decoder_num_stacks == 0:
            # Keep track of the backbone output, since it might be used as the
            # semantic feature output.
            backbone_output = activated_output
            # Now that we have finished building the backbone, we either return the
            # classification logits, or continue building a non-backbone decoder for
            # panoptic segmentation.
            if self._classification_mode:
                endpoints['backbone_output'] = backbone_output
                return None, None, None, endpoints
            else:
                current_is_backbone = False

        if not current_is_backbone:
            # Build extra layers if we have finished building the backbone.
            current_name = '_stage5_' + EXTRA
            current_output, memory_feature = (
                getattr(self, current_name)(
                    (current_output, memory_feature), training=training))

        # Compute parameter lists for stacked decoder.
        total_decoder_num_stacks = (
                self._backbone_decoder_num_stacks + self._extra_decoder_num_stacks)

        # Keep track of all endpoints that will be used in the stacked decoder.
        stride_to_features = {}
        stride_to_features[min(2, self._output_stride)] = [endpoints['stage1']]
        stride_to_features[min(4, self._output_stride)] = [endpoints['stage2']]
        stride_to_features[min(8, self._output_stride)] = [endpoints['stage3']]
        stride_to_features[min(16, self._output_stride)] = [endpoints['stage4']]
        # Only keep the last endpoint from the backbone with the same resolution,
        # i.e., if the output stride is 16, the current output will override
        # the stride 16 endpoint, endpoints['res4'].
        stride_to_features[min(32, self._output_stride)] = [current_output]

        # Use a function to compute the next stride.
        def next_stride_fn(x):
            return x // 2

        current_decoder_stride = self._output_stride
        decoder_stage = 0

        # Exit if we have enough stacks and reach the decoding output stride.
        while (current_stack < total_decoder_num_stacks or
               current_decoder_stride > self._high_resolution_output_stride):
            decoder_stage += 1
            current_decoder_stride = next_stride_fn(current_decoder_stride)

            if current_decoder_stride == self._output_stride:
                current_stack += 1
                # Switch the decoder direction if we reach the largest stride.

                def next_stride_fn(x):
                    return x // 2

            # Include the current feature and two previous features from the target
            # resolution in the decoder. We select two because it contains one upward
            # feature and one downward feature, but better choices are possible.
            decoder_features_list = (
                    [current_output] +
                    stride_to_features[current_decoder_stride][-2:])

            # Fuse and resize features with striding, resizing and 1x1 convolutions.
            if current_is_backbone:
                current_name = '_decoder_stage{}_resized_fuse'.format(decoder_stage)
            else:
                current_name = '_decoder_stage{}_{}_resized_fuse'.format(
                    decoder_stage, EXTRA)
            current_output = getattr(self, current_name)(
                decoder_features_list, training=training)

            # Apply a decoder block group for building the backbone.
            if current_is_backbone:
                current_name = '_decoder_stage{}'.format(decoder_stage)
                current_output, memory_feature = (
                    getattr(self, current_name)(
                        (current_output, memory_feature), training=training))

            if (current_decoder_stride == self._output_stride and
                    current_stack == self._backbone_decoder_num_stacks):
                # Keep track of the backbone output, since it might be used as the
                # semantic feature output.
                activated_output = self._activation_fn(current_output)
                backbone_output = activated_output
                # Now that we have finished building the backbone, we either return the
                # classification logits, or continue building a non-backbone decoder for
                # panoptic segmentation.
                if self._classification_mode:
                    endpoints['backbone_output'] = backbone_output
                    return None, None, None, endpoints
                else:
                    current_is_backbone = False

            # Apply a decoder block group for building the extra layers.
            if not current_is_backbone:
                current_name = '_decoder_stage{}_{}'.format(decoder_stage, EXTRA)
                current_output, memory_feature = (
                    getattr(self, current_name)(
                        (current_output, memory_feature), training=training))

            # Append the current feature into the feature dict for possible later
            # usage.
            stride_to_features[current_decoder_stride].append(current_output)
            if current_decoder_stride == self._high_resolution_output_stride:
                activated_output = self._activation_fn(current_output)
                high_resolution_outputs.append(activated_output)

                def next_stride_fn(x):
                    return x * 2
        return memory_feature, high_resolution_outputs, backbone_output, endpoints

    def call_extra_endpoints(self,
                             memory_feature,
                             high_resolution_outputs,
                             backbone_output,
                             endpoints,
                             training=False):
        """Performs a forward pass to generate extra endpoints.

        Args:
          memory_feature: None if no transformer is used. A [batch, num_memory,
            memory_channel] tensor if transformer is used.
          high_resolution_outputs: A list of decoded tensors with
            high_resolution_output_stride.
          backbone_output: An output tensor of the backbone, with output_stride.
          endpoints: A dict, the network endpoints that might be used by DeepLab.
          training: A boolean, whether the model is in training mode.

        Returns:
          endpoints: A dict, the network endpoints that might be used by DeepLab.
        """
        # Assert that we have already returned if we are building a classifier.
        assert not self._classification_mode
        if (self._backbone_use_transformer_beyond_stride or
                self._extra_decoder_use_transformer_beyond_stride):
            # Build extra memory path feed forward networks for the class feature and
            # the mask feature.
            class_feature = getattr(self, '_class_feature_' + EXTRA)(
                memory_feature, training=training)
            mask_feature = getattr(self, '_mask_feature_' + EXTRA)(
                memory_feature, training=training)
            endpoints['transformer_class_feature'] = class_feature
            endpoints['transformer_mask_feature'] = mask_feature

        # Output the last high resolution feature as panoptic feature.
        endpoints['feature_panoptic'] = high_resolution_outputs[-1]

        # Avoid sharing our panoptic feature with the semantic auxiliary loss. So we
        # use the backbone feature or the decoded backbone feature for the semantic
        # segmentation head (i.e. the auxiliary loss).
        if self._extra_decoder_num_stacks:
            endpoints['feature_semantic'] = (
                high_resolution_outputs[self._backbone_decoder_num_stacks])
        else:
            endpoints['feature_semantic'] = backbone_output
        endpoints['backbone_output'] = backbone_output
        return endpoints

    def call(self, inputs, training=False):
        """Performs a forward pass.

        Args:
          inputs: An input [batch, height, width, channel] tensor.
          training: A boolean, whether the model is in training mode.

        Returns:
          endpoints: A dict, the network endpoints that might be used by DeepLab.
        """
        current_output, activated_output, memory_feature, endpoints = (
            self.call_encoder_before_stacked_decoder(inputs, training=training))
        memory_feature, high_resolution_outputs, backbone_output, endpoints = (
            self.call_stacked_decoder(current_output,
                                      activated_output,
                                      memory_feature,
                                      endpoints,
                                      training=training))
        if self._classification_mode:
            return endpoints
        endpoints = self.call_extra_endpoints(memory_feature,
                                              high_resolution_outputs,
                                              backbone_output,
                                              endpoints,
                                              training=training)
        return endpoints



#!/usr/bin/env python3
"""
TODO: rectify with run_teamfeat_acsc_landcover.py
"""
import ubelt as ub
import scriptconfig as scfg
from geowatch.cli.smartflow_ingress import smartflow_ingress
from geowatch.cli.smartflow_egress import smartflow_egress


class TeamFeatLandcoverConfig(scfg.DataConfig):
    """
    Run DZYNE landcover feature computation as baseline framework component
    """
    input_path = scfg.Value(None, type=str, position=1, required=True, help=ub.paragraph(
            '''
            Path to input T&E Baseline Framework JSON
            '''))
    input_region_path = scfg.Value(None, type=str, position=2, required=True, help=ub.paragraph(
            '''
            Path to input T&E Baseline Framework Region definition JSON
            '''))
    output_path = scfg.Value(None, type=str, position=3, required=True, help='S3 path for output JSON')
    model_path = scfg.Value(None, type=str, required=True, help='File path to landcover model')
    aws_profile = scfg.Value(None, type=str, help=ub.paragraph(
            '''
            AWS Profile to use for AWS S3 CLI commands
            '''))
    dryrun = scfg.Value(False, isflag=True, short_alias=['d'], help='Run AWS CLI commands with --dryrun flag')
    outbucket = scfg.Value(None, type=str, required=True, short_alias=['o'], help=ub.paragraph(
            '''
            S3 Output directory for STAC item / asset egress
            '''))
    requester_pays = scfg.Value(False, isflag=True, short_alias=['r'], help=ub.paragraph(
            '''
            Run AWS CLI commands with `--requestor_payer requester` flag
            '''))
    newline = scfg.Value(False, isflag=True, short_alias=['n'], help=ub.paragraph(
            '''
            Output as simple newline separated STAC items
            '''))


def main():
    config = TeamFeatLandcoverConfig.cli(strict=True)
    print('config = {}'.format(ub.urepr(dict(config), nl=1, align=':')))
    run_landcover_for_baseline(config)


def run_landcover_for_baseline(config):
    from geowatch.utils.util_framework import download_region

    ####
    # DEBUGGING:
    # Print info about what version of the code we are running on
    from geowatch.utils.util_framework import NodeStateDebugger
    node_state = NodeStateDebugger()
    node_state.print_environment()
    node_state.print_local_invocation(config)

    # 1. Ingress data
    print("* Running baseline framework kwcoco ingress *")
    ingress_dir = ub.Path('/tmp/ingress')
    ingressed_assets = smartflow_ingress(
        config.input_path,
        ['enriched_bas_kwcoco_file',
         'enriched_bas_kwcoco_teamfeats',
         'enriched_bas_kwcoco_rawbands'],
        ingress_dir,
        config.aws_profile,
        config.dryrun)

    # 2. Download and prune region file
    print("* Downloading and pruning region file *")
    local_region_path = '/tmp/region.json'
    download_region(
        input_region_path=config.input_region_path,
        output_region_path=local_region_path,
        aws_profile=config.aws_profile,
        strip_nonregions=True,
    )

    # 2. Generate Landcover features
    print("* Generating landcover features *")
    dzyne_landcover_features_kwcoco_path = ingress_dir / 'dzyne_landcover_kwcoco.json'

    enriched_bas_kwcoco_file = ingressed_assets['enriched_bas_kwcoco_file']
    ub.cmd([
        'python', '-m', 'geowatch.tasks.landcover.predict',
        '--dataset', enriched_bas_kwcoco_file,
        '--deployed', config.model_path,
        '--output', dzyne_landcover_features_kwcoco_path,
        '--num_workers', '2',
        '--with_hidden', '32',
        '--assets_dname', '_teamfeats',
        '--select_images', '.sensor_coarse == "S2"',
        '--device', '0'
    ], check=True, verbose=3, capture=False)

    # 3. Combining landcover features with input features to pass to BAS
    print("* Combining input features with computed landcover features *")
    combo_features_kwcoco_path = ingress_dir / 'features_combo_with_landcover_kwcoco.json'
    ub.cmd([
        'python', '-m', 'geowatch.cli.coco_combine_features',
        '--src', enriched_bas_kwcoco_file, dzyne_landcover_features_kwcoco_path,
        '--dst', combo_features_kwcoco_path,
    ], check=True, verbose=3, capture=False)

    # Reroot kwcoco files to make downloaded results easier to work with
    ub.cmd(['kwcoco', 'reroot', f'--src={combo_features_kwcoco_path}', '--inplace=1', '--absolute=0'])

    # 4. Egress (envelop KWCOCO dataset in a STAC item and egress;
    #    will need to recursive copy the kwcoco output directory up to
    #    S3 bucket)
    print("* Egressing KWCOCO dataset and associated STAC item *")
    # Add new assets to be egressed
    ingressed_assets['timecombined_kwcoco_file_for_bas_with_landcover'] = (
        combo_features_kwcoco_path)
    ingressed_assets['landcover_assets'] = ingress_dir / '_teamfeats'

    ingressed_assets['enriched_bas_kwcoco_file'] = combo_features_kwcoco_path
    ingressed_assets['enriched_bas_kwcoco_teamfeats'] = ingress_dir / '_teamfeats'

    smartflow_egress(ingressed_assets,
                     local_region_path,
                     config.output_path,
                     config.outbucket,
                     aws_profile=None,
                     dryrun=False,
                     newline=False)


if __name__ == "__main__":
    main()



from typing import TYPE_CHECKING

__all__ = [
    "_ArgumentGroup",
    "ArgumentParser",
    "ruyamlCommentedMap",
]

if TYPE_CHECKING:  # pragma: no cover
    from ruyaml.comments import CommentedMap as ruyamlCommentedMap

    from ._core import ArgumentParser, _ArgumentGroup
else:
    globals().update({k: None for k in __all__})



#!/usr/bin/env python3
"""
See Old Script:
    ~/code/watch/scripts/run_stac_to_cropped_kwcoco.py
"""
import sys
import traceback
import os
import json
import shutil

from geowatch.cli.baseline_framework_ingress import baseline_framework_ingress, load_input_stac_items
from geowatch.cli.smartflow_egress import smartflow_egress
from geowatch.cli.stac_to_kwcoco import stac_to_kwcoco
from geowatch.cli import coco_add_watch_fields
from geowatch.utils.util_framework import download_region
import ubelt as ub
import scriptconfig as scfg


class BASDatasetConfig(scfg.DataConfig):
    """
    Generate cropped KWCOCO dataset for BAS from STAC
    """
    input_path = scfg.Value(None, type=str, position=1, required=True, help=ub.paragraph(
        '''
        Path to the STAC items this step can use as inputs.
        This is usually an S3 Path.
        '''), alias=['input_stac_path'])

    input_region_path = scfg.Value(None, type=str, position=2, required=True, help=ub.paragraph(
        '''
        Path to input T&E Baseline Framework Region definition JSON
        '''))

    output_path = scfg.Value(None, type=str, position=3, required=True, help=ub.paragraph(
        '''
        Path to the STAC items that register the outputs of this stage.
        This is usually an S3 Path.
        '''), alias=['output_stac_path'])

    from_collated = scfg.Value(False, isflag=True, help=ub.paragraph(
        '''
        Data to convert has been run through TA-1 collation
        '''))

    aws_profile = scfg.Value(None, type=str, help=ub.paragraph(
        '''
        AWS Profile to use for AWS S3 CLI commands
        '''))

    dryrun = scfg.Value(False, isflag=True, short_alias=['d'], help='Run AWS CLI commands with --dryrun flag')

    virtual = scfg.Value(False, isflag=True, help=ub.paragraph(
        '''
        Ingress will be virtual (using GDAL's virtual file system)
        '''))

    outbucket = scfg.Value(None, type=str, required=True, short_alias=['o'], help=ub.paragraph(
        '''
        S3 Output directory for STAC item / asset egress
        '''))

    requester_pays = scfg.Value(False, isflag=True, short_alias=['r'], help=ub.paragraph(
        '''
        Run AWS CLI commands with `--requestor_payer requester` flag
        '''))

    newline = scfg.Value(False, isflag=True, short_alias=['n'], help=ub.paragraph(
        '''
        Output as simple newline separated STAC items
        '''))

    jobs = scfg.Value(1, type=int, short_alias=['j'], help='Number of jobs to run in parallel')

    dont_recompute = scfg.Value(False, isflag=True, help=ub.paragraph(
        '''
        Will not recompute if output_path already exists
        '''))

    previous_interval_output = scfg.Value(None, type=str, help=ub.paragraph(
        '''
        Output path for previous interval BAS DatasetGen step
        '''))

    bas_align_config = scfg.Value(None, type=str, help=ub.paragraph(
        '''
        The configuration for the coco-align step
        '''))

    time_combine_config = scfg.Value(None, type=str, help=ub.paragraph(
        '''
        If specified, perform time combine on BAS dataset. The special key
        enabled will disable the computation.
        '''), alias=['time_combine'])

    skip_timecombine_on_fail = scfg.Value(False, help=ub.paragraph(
        '''
        If an error occurs during the timecombine call, output empty KWCOCO.
        '''))

    def __post_init__(self):
        if self.time_combine_config in {False, None, 'False', 'None'}:
            self.time_combine_config = {'enabled': False}
        elif self.time_combine_config in {True, 'True'}:
            self.time_combine_config = {'enabled': True}


def main():
    config = BASDatasetConfig.cli(strict=True)
    print('config = {}'.format(ub.urepr(config, nl=1, align=':')))
    run_stac_to_cropped_kwcoco(config)


def build_combined_stac(previous_stac_input_path,
                        stac_input_path,
                        combined_stac_output_path):
    previous_stac_items = load_input_stac_items(previous_stac_input_path, None)
    current_stac_items = load_input_stac_items(stac_input_path, None)

    combined_stac_items = previous_stac_items.copy()
    combined_stac_items.extend(current_stac_items)

    with open(combined_stac_output_path, 'w') as f:
        print('\n'.join((json.dumps(item)
                         for item in combined_stac_items)), file=f)

    return combined_stac_output_path


def input_stac_to_kwcoco(stac_items_path,
                         working_dir,
                         out_kwcoco_filename,
                         target_gsd,
                         aws_profile=None,
                         dryrun=False,
                         requester_pays=False,
                         jobs=1,
                         virtual=False,
                         from_collated=False):
    working_dir = ub.Path(working_dir)
    os.makedirs(working_dir, exist_ok=True)
    print("* Running baseline framework ingress *")
    ingressed_catalog = baseline_framework_ingress(
        stac_items_path,
        working_dir,
        aws_profile=aws_profile,
        dryrun=dryrun,
        requester_pays=requester_pays,
        relative=False,
        jobs=jobs,
        virtual=virtual)

    # 3. Convert ingressed STAC catalog to KWCOCO
    print("* Converting STAC to KWCOCO *")
    out_kwcoco_path = working_dir / out_kwcoco_filename
    stac_to_kwcoco(ingressed_catalog,
                   out_kwcoco_path,
                   assume_relative=False,
                   populate_watch_fields=False,
                   jobs=jobs,
                   from_collated=from_collated,
                   ignore_duplicates=True)

    print("* Adding geowatch feilds *")
    coco_add_watch_fields.main(cmdline=False,
                               src=out_kwcoco_path,
                               inplace=True,
                               target_gsd=target_gsd,
                               workers=jobs)

    return out_kwcoco_path


def run_stac_to_cropped_kwcoco(config):
    from geowatch.utils import util_fsspec
    from kwutil.util_yaml import Yaml
    from delayed_image.channel_spec import ChannelSpec
    # from kwcoco import ChannelSpec
    from geowatch.cli import coco_align
    from geowatch.cli import coco_time_combine
    from geowatch.mlops.pipeline_nodes import ProcessNode
    from geowatch.cli.smartflow_ingress import smartflow_ingress
    import kwcoco

    from geowatch.utils.util_framework import NodeStateDebugger
    node_state = NodeStateDebugger()
    node_state.print_environment()
    node_state.print_local_invocation(config)

    if config.aws_profile is not None:
        # This should be sufficient, but it is not tested.
        util_fsspec.S3Path._new_fs(profile=config.aws_profile)

    align_config_default = ub.udict(Yaml.coerce(ub.codeblock(
        f'''
        force_nodata: -9999
        include_channels: "coastal|blue|green|red|B05|B06|B07|nir|B8A|B09|cirrus|swir16|swir22|pan|quality"
        geo_preprop: auto
        keep: null
        convexify_regions: True
        target_gsd: 10GSD
        context_factor: 1
        force_min_gsd: 10
        img_workers: {str(config.jobs)}
        aux_workers: auto
        rpc_align_method: affine_warp
        ''')))

    time_combine_config_default = ub.udict(Yaml.coerce(ub.codeblock(
        '''
        enabled: True
        time_window: '1y'
        channels: 'coastal|blue|green|red|B05|B06|B07|nir|B8A|B09|cirrus|swir16|swir22|pan'
        resolution: '10GSD'
        workers: 'avail'
        start_time: '1970-01-01'
        merge_method: 'mean'
        assets_dname: 'raw_bands'
        ''')))

    align_config = align_config_default | Yaml.coerce(config.bas_align_config)
    if align_config['aux_workers'] == 'auto':
        # Auto gives each channel its own aux worker
        num_channels = align_config['include_channels'].count('|') + 1
        align_config['aux_workers'] = num_channels

    time_combine_config = time_combine_config_default | Yaml.coerce(config.time_combine_config)

    if time_combine_config['channels'] == 'auto':
        # Default time combine channels to the align channels minus quality.
        time_combine_config['channels'] = ChannelSpec.coerce(align_config['include_channels']) - {'quality'}

    time_combine_enabled = time_combine_config.pop('enabled', True)

    target_gsd = align_config['target_gsd']

    if config.dont_recompute:
        output_path = util_fsspec.FSPath.coerce(config.output_path)
        if output_path.exists():
            print('Dont recompute is True. Early stopping')
            # If output_path file was there, nothing to do
            return

    # `ta1_cropped_dir` is the directory that gets recursively copied
    # up to S3, want to put any kwcoco manifests we may need
    # downstream into this directory.  TODO: rename variable to
    # include something like upload_dir or output_dir
    ta1_cropped_dir = ub.Path('/tmp/cropped_kwcoco')
    local_region_path = ub.Path('/tmp/region.json')
    ingress_dir = ub.Path('/tmp/ingress')

    ta1_cropped_dir.ensuredir()
    ingress_dir.ensuredir()

    ta1_kwcoco_path = ingress_dir / 'ingress_kwcoco.json'
    ta1_bas_kwcoco_path = ta1_cropped_dir / 'kwcoco_for_bas.json'
    ta1_cropped_kwcoco_path = ta1_cropped_dir / 'cropped_kwcoco_for_bas.json'

    align_config['src'] = ta1_bas_kwcoco_path
    align_config['dst'] = ta1_cropped_kwcoco_path
    align_config['regions'] = local_region_path
    # Validate config before running stuff
    align_config = coco_align.CocoAlignGeotiffConfig(**align_config)
    print('align_config = {}'.format(ub.urepr(align_config, nl=1)))

    if time_combine_enabled:
        preproc_kwcoco_fpath = ub.Path(ta1_cropped_kwcoco_path).augment(
            stemsuffix='_timecombined', ext='.kwcoco.zip', multidot=True)
        time_combine_config['input_kwcoco_fpath'] = ta1_cropped_kwcoco_path
        time_combine_config['output_kwcoco_fpath'] = preproc_kwcoco_fpath
        # Validate config before running stuff
        time_combine_config = coco_time_combine.TimeCombineConfig(**time_combine_config)
        print('time_combine_config = {}'.format(ub.urepr(time_combine_config, nl=1)))

    # 2. Download and prune region file
    print("* Downloading and pruning region file *")
    local_region_path = download_region(config.input_region_path,
                                        local_region_path,
                                        aws_profile=config.aws_profile,
                                        strip_nonregions=True)

    # HACK: this is what coco-align outputs by default. We should have this be
    # explicit and configurable so we can set it to what we want here.
    from geowatch.utils.util_framework import determine_region_id
    region_id = determine_region_id(local_region_path)
    ta1_cropped_rawband_dpath = ta1_cropped_dir / region_id

    node_state.print_current_state(ingress_dir)

    from geowatch.geoannots.geomodels import RegionModel
    region = RegionModel.coerce(local_region_path)

    # Returned as datetime
    current_interval_end_date = region.end_date

    if current_interval_end_date.month == 1 and current_interval_end_date.day == 1:
        # If current interval ends at the start of a year,
        # consider the "current" year to be the previous one
        current_interval_year = current_interval_end_date.year - 1
    else:
        current_interval_year = current_interval_end_date.year

    # Download STAC input file locally
    local_stac_path = ingress_dir / 'input_stac.jsonl'
    input_stac_path = util_fsspec.FSPath.coerce(config.input_path)
    input_stac_path.copy(util_fsspec.FSPath.coerce(local_stac_path))

    # 3. Generate KWCOCO dataset from input STAC
    current_interval_kwcoco_path = input_stac_to_kwcoco(
        config.input_path,
        ingress_dir,
        ta1_kwcoco_path,
        target_gsd,
        aws_profile=config.aws_profile,
        dryrun=config.dryrun,
        requester_pays=config.requester_pays,
        jobs=config.jobs,
        virtual=config.virtual,
        from_collated=config.from_collated)

    # Overwritten if incremental mode (and not first interval)
    combined_kwcoco_path = current_interval_kwcoco_path

    # Setting 'combined_stac_input' here to ensure we have something
    # from the first interval.  Gets overwritten if not the first
    # interval
    incremental_assets_for_egress = {'combined_stac_input': local_stac_path}
    previous_ingressed_assets = None
    if config.previous_interval_output is not None:
        print('* Combining previous interval time combined kwcoco with '
              'current *')
        previous_ingress_dir = ub.Path('/tmp/ingress_previous')
        try:
            previous_ingressed_assets = smartflow_ingress(
                config.previous_interval_output,
                ['combined_stac_input',
                 'timecombined_kwcoco_file_for_bas',
                 'timecombined_kwcoco_file_for_bas_assets'],
                previous_ingress_dir,
                config.aws_profile,
                config.dryrun)
        except FileNotFoundError:
            print("** Warning: Couldn't ingress previous interval output; "
                  "assuming this is the first interval **")
        else:
            combined_stac_path = ingress_dir / 'combined_stac_items.jsonl'

            build_combined_stac(
                previous_ingressed_assets['combined_stac_input'],
                local_stac_path,
                combined_stac_path)

            incremental_assets_for_egress['combined_stac_input'] =\
                combined_stac_path

            # Perform the filtering
            selected_stac_items = []
            with open(combined_stac_path) as f:
                for line in f:
                    try:
                        stac_item = json.loads(line)
                    except json.decoder.JSONDecodeError:
                        print("** Warning: Couldn't parse STAC item from "
                              "'combined_stac_input', skipping!")
                        continue

                    if stac_item['properties']['datetime'].startswith(
                            str(current_interval_year)):
                        selected_stac_items.append(stac_item)

            filtered_stac_items_path = ingress_dir / 'filtered_stac_items.jsonl'
            with open(filtered_stac_items_path, 'w') as f:
                print('\n'.join((json.dumps(item)
                                 for item in selected_stac_items)), file=f)

            incremental_assets_for_egress['filtered_stac_input'] =\
                filtered_stac_items_path

            combined_kwcoco_path = ingress_dir / 'combined_timecombined_kwcoco.json'

            combined_kwcoco_path = input_stac_to_kwcoco(
                filtered_stac_items_path,
                ingress_dir,
                combined_kwcoco_path,
                target_gsd,
                aws_profile=config.aws_profile,
                dryrun=config.dryrun,
                requester_pays=config.requester_pays,
                jobs=config.jobs,
                virtual=config.virtual,
                from_collated=config.from_collated)

    # 3a. Filter KWCOCO dataset by sensors used for BAS
    # Will use either the combined KWCOCO dataset (for incremental
    # mode) or strictly the input STAC items
    print("* Filtering KWCOCO dataset for BAS")
    ub.cmd([
        'kwcoco', 'subset',
        '--src', combined_kwcoco_path,
        '--dst', ta1_bas_kwcoco_path,
        '--absolute', 'False',
        # '--select_images',
        # '.sensor_coarse == "L8" or .sensor_coarse == "S2"'
    ], check=True, verbose=3, capture=False)

    # 3b. Filter KWCOCO dataset by sensors used for SC
    # TODO: move this to run_sc_datagen
    print("* Filtering KWCOCO dataset for SC")
    ta1_sc_kwcoco_path = ta1_cropped_dir / 'kwcoco_for_sc.json'
    ub.cmd(['kwcoco', 'subset',
            '--src', ta1_kwcoco_path,
            '--dst', ta1_sc_kwcoco_path,
            '--absolute', 'False',
            # '--select_images',
            # '.sensor_coarse == "WV1" or .sensor_coarse == "WV" or .sensor_coarse == "S2"'
            ],
           check=True, verbose=3, capture=False)

    # 4. Crop ingress KWCOCO dataset to region for BAS
    print("* Cropping KWCOCO dataset to region for BAS*")

    ALIGN_EXEC_MODE = 'cmd'
    # Not sure if one is more stable than the other
    if ALIGN_EXEC_MODE == 'import':
        coco_align.main(cmdline=False, **align_config)
    elif ALIGN_EXEC_MODE == 'cmd':
        align_node = ProcessNode(
            command='python -m geowatch.cli.coco_align',
            config=align_config,
        )
        command = align_node.final_command()
        ub.cmd(command, check=True, capture=False, verbose=3, shell=True)
    else:
        raise KeyError(ALIGN_EXEC_MODE)

    ### Filter / clean geotiffs (probably should be a separate step)
    CLEAN_GEOTIFFS = 0
    if CLEAN_GEOTIFFS:
        # Detect blocky black regions in geotiffs and switch them to NODATA
        # Modifies geotiffs inplace
        remove_bad_images_node = ProcessNode(
            command='geowatch clean_geotiffs',
            in_paths={
                'src': ta1_cropped_kwcoco_path,
            },
            config={
                'prefilter_channels': 'red',
                'channels': 'red|green|blue|nir',
                'workers': 'avail',
                'dry': False,
                'probe_scale': None,
                'nodata_value': -9999,
                'min_region_size': 256,
            },
            node_dpath='.'
        )
        command = remove_bad_images_node.final_command()
        ub.cmd(command, shell=True, capture=False, verbose=3, check=True)

    REMOVE_BAD_IMAGES = 0
    if REMOVE_BAD_IMAGES:
        # Remove images that are nearly all nan
        remove_bad_images_node = ProcessNode(
            command='geowatch remove_bad_images',
            in_paths={
                'src': ta1_cropped_kwcoco_path,
            },
            out_paths={
                'dst': ta1_cropped_kwcoco_path,  # hack: this is inplace, fix it if we enable.
            },
            config={
                'workers': 'avail',
                'interactive': False,
                'overview': 0,
            },
            node_dpath='.'
        )
        command = remove_bad_images_node.final_command()
        ub.cmd(command, shell=True, capture=False, verbose=3, check=True)
    else:
        print('Not removing bad images. TODO: add support')
        # ta1_sc_cropped_kwcoco_prefilter_path.copy(ta1_sc_cropped_kwcoco_path)

    # Reroot the kwcoco files to be relative and make it easier to work with
    # downloaded results
    ub.cmd(['kwcoco', 'reroot', f'--src={ta1_cropped_kwcoco_path}', '--inplace=1', '--absolute=0'])

    node_state.print_current_state(ingress_dir)

    # 5. Do the time_combine for BAS
    if time_combine_enabled:
        try:
            coco_time_combine.main(
                cmdline=0,
                **time_combine_config
            )
        except Exception:
            if config.skip_timecombine_on_fail:
                print("WARNING: Exception occurred (printed below), generating empty KWCOCO for time-combined output")
                traceback.print_exception(*sys.exc_info())

                empty_dset_path = ta1_cropped_dir / 'empty.json'
                empty_dset = kwcoco.CocoDataset()
                empty_dset.dump(empty_dset_path)

                final_interval_bas_kwcoco_path = empty_dset_path
            else:
                raise
        else:
            # Add geowatch feilds
            print("* Adding geowatch feilds to time combined data *")
            coco_add_watch_fields.main(cmdline=False,
                                       src=preproc_kwcoco_fpath,
                                       dst=preproc_kwcoco_fpath,
                                       target_gsd=target_gsd,
                                       workers=config.jobs)
            final_interval_bas_kwcoco_path = preproc_kwcoco_fpath

        # Reroot the kwcoco files to be relative and make it easier to work with
        # downloaded results
        ub.cmd(['kwcoco', 'reroot', f'--src={final_interval_bas_kwcoco_path}', '--inplace=1', '--absolute=0'])

    else:
        final_interval_bas_kwcoco_path = ta1_cropped_kwcoco_path

    # 6.1. Combine previous interval time-combined data for BAS
    if config.previous_interval_output is not None and previous_ingressed_assets is not None:
        combined_timecombined_kwcoco_path =\
            ta1_cropped_dir / 'combined_timecombined_kwcoco.json'

        previous_timecombined_kwcoco_path = ub.Path(
            previous_ingressed_assets['timecombined_kwcoco_file_for_bas'])

        import kwcoco
        previous_timecombined_dset = kwcoco.CocoDataset(
            previous_timecombined_kwcoco_path)

        image_ids_to_remove =\
            [o["id"] for o in previous_timecombined_dset.images().objs
             if o['date_captured'].startswith(str(current_interval_year))]

        previous_timecombined_dset.remove_images(image_ids_to_remove)

        filtered_previous_timecombined_kwcoco_path =\
            previous_ingress_dir / 'filtered_combined_timecombined_kwcoco.json'

        incremental_assets_for_egress['filtered_combined_timecombined_kwcoco'] =\
            filtered_previous_timecombined_kwcoco_path

        previous_timecombined_dset.dump(
            filtered_previous_timecombined_kwcoco_path)

        # On first interval nothing will be copied down so need to
        # check that we have the input explicitly
        from geowatch.cli.concat_kwcoco_videos import concat_kwcoco_datasets
        if filtered_previous_timecombined_kwcoco_path.is_file() and len(previous_timecombined_dset.images()) > 0:
            # Don't bother to concatenate if previous (now filtered)
            # dset is empty (has no images)
            concat_kwcoco_datasets(
                (filtered_previous_timecombined_kwcoco_path,
                 final_interval_bas_kwcoco_path),
                combined_timecombined_kwcoco_path)
            # Copy saliency assets from previous bas fusion
            shutil.copytree(
                previous_ingress_dir / 'raw_bands',
                ta1_cropped_dir / 'raw_bands',
                dirs_exist_ok=True)
        else:
            # This is either the first interval, or previous
            # interval(s) only contain images from the current
            # interval year
            combined_timecombined_kwcoco_path = final_interval_bas_kwcoco_path
    else:
        combined_timecombined_kwcoco_path = final_interval_bas_kwcoco_path

    # 7. Egress (envelop KWCOCO dataset in a STAC item and egress;
    #    will need to recursive copy the kwcoco output directory up to
    #    S3 bucket)

    timecombined_rawband_dpath = ta1_cropped_dir / 'raw_bands'
    timecombined_teamfeat_dpath = ta1_cropped_dir / '_teamfeats'
    # Put a dummy file in these directories so we can upload a nearly-empty folder
    # to S3
    timecombined_rawband_dpath.ensuredir()
    (timecombined_rawband_dpath / 'dummy').write_text('dummy')
    timecombined_teamfeat_dpath.ensuredir()
    (timecombined_teamfeat_dpath / 'dummy').write_text('dummy')
    ta1_cropped_rawband_dpath.ensuredir()
    (ta1_cropped_rawband_dpath / 'dummy').write_text('dummy')

    node_state.print_current_state(ingress_dir)

    if 1:
        # Print debug info about items we are about to egress
        print('-- stats about combined_timecombined_kwcoco_path --')
        ub.cmd(f'kwcoco stats {combined_timecombined_kwcoco_path}', verbose=3)
        ub.cmd(f'geowatch stats {combined_timecombined_kwcoco_path}', verbose=3)
        print('-- stats about ta1_sc_kwcoco_path --')
        ub.cmd(f'kwcoco stats {ta1_sc_kwcoco_path}', verbose=3)
        ub.cmd(f'geowatch stats {ta1_sc_kwcoco_path}', verbose=3)

    print("* Egressing KWCOCO dataset and associated STAC item *")
    assets_to_egress = {
        'timecombined_kwcoco_file_for_bas': combined_timecombined_kwcoco_path,
        'timecombined_kwcoco_file_for_bas_assets': timecombined_rawband_dpath,

        # This is an alias to the BAS dataset and assets that team feature
        # scripts will update.
        'enriched_bas_kwcoco_file': combined_timecombined_kwcoco_path,
        'enriched_bas_kwcoco_teamfeats': timecombined_teamfeat_dpath,
        'enriched_bas_kwcoco_rawbands': timecombined_rawband_dpath,

        # TODO: @DMJ: I dont think anything uses this? Can it be removed?
        # JPC: Seems like the answer is no, for now. I've seen this used later
        # on in the sc datagen node, although the stac-to-kwcoco for sc could
        # be run there.
        'kwcoco_for_sc': ta1_sc_kwcoco_path,

        # We need to egress the temporally dense dataset for COLD
        'timedense_bas_kwcoco_file': ta1_cropped_kwcoco_path,
        'timedense_bas_kwcoco_rawbands': ta1_cropped_rawband_dpath,
        **incremental_assets_for_egress
    }
    smartflow_egress(assets_to_egress,
                     local_region_path,
                     config.output_path,
                     config.outbucket,
                     aws_profile=config.aws_profile,
                     dryrun=config.dryrun,
                     newline=config.newline)


if __name__ == "__main__":
    main()



import skimage.io
from functools import partial
from multiprocess import Pool
import itertools

import numpy as np
import rasterio
import rasterio.rio.insp
import fiona
from pyproj import Proj, transform

import torch_lydorn.torch.utils.data

import os

from tqdm import tqdm

import torch

from frame_field_learning import data_transforms

from lydorn_utils import run_utils
from lydorn_utils import print_utils
from lydorn_utils import python_utils
from lydorn_utils import polygon_utils
from lydorn_utils import ogr2ogr
from lydorn_utils import image_utils


class LuxcartaBuildings(torch_lydorn.torch.utils.data.Dataset):
    def __init__(self, root, transform=None, pre_transform=None, fold="train", patch_size=None, patch_stride=None,
                 pool_size=1):
        assert fold in {"train", "test"}, "fold should be either train of test"
        self.fold = fold
        self.patch_size = patch_size
        self.patch_stride = patch_stride
        self.pool_size = pool_size
        self._processed_filepaths = None
        # TODO: implement pool_size option
        super(LuxcartaBuildings, self).__init__(root, transform, pre_transform)

    @property
    def raw_dir(self):
        return os.path.join(self.root, 'raw', self.fold)

    @property
    def processed_dir(self):
        return os.path.join(self.root, 'processed', self.fold)

    @property
    def raw_file_names(self):
        return []

    @property
    def raw_sample_metadata_list(self):
        returned_list = []
        for dirname in os.listdir(self.raw_dir):
            dirpath = os.path.join(self.raw_dir, dirname)
            if os.path.isdir(dirpath):
                image_filepath_list = python_utils.get_filepaths(dirpath, endswith_str="_crop.tif",
                                                                 startswith_str="Ortho",
                                                                 not_endswith_str=".tif_crop.tif")
                gt_polygons_filepath_list = python_utils.get_filepaths(dirpath, endswith_str=".shp",
                                                                       startswith_str="Building")
                mask_filepath_list = python_utils.get_filepaths(dirpath, endswith_str=".kml", startswith_str="zone")
                if len(image_filepath_list) and len(gt_polygons_filepath_list):
                    metadata = {
                        "dirname": dirname,
                        "image_filepath": image_filepath_list[0],
                        "gt_polygons_filepath": gt_polygons_filepath_list[0],
                    }
                    if len(mask_filepath_list):
                        metadata["mask_filepath"] = mask_filepath_list[0]
                    returned_list.append(metadata)
        return returned_list

    @property
    def processed_filepaths(self):
        if self._processed_filepaths is None:
            self._processed_filepaths = []
            for dirname in os.listdir(self.processed_dir):
                dirpath = os.path.join(self.processed_dir, dirname)
                if os.path.isdir(dirpath):
                    filepath_list = python_utils.get_filepaths(dirpath, endswith_str=".pt", startswith_str="data.")
                    self._processed_filepaths.extend(filepath_list)
        return self._processed_filepaths

    def __len__(self):
        return len(self.processed_filepaths)

    def _download(self):
        pass

    def download(self):
        pass

    def _get_mask_multi_polygon(self, mask_filepath, shp_srs):
        # Read mask and convert to shapefile's crs:
        mask_shp_filepath = os.path.join(os.path.dirname(mask_filepath), "mask.shp")

        # TODO: see what's up with the warnings:
        ogr2ogr.main(["", "-f", "ESRI Shapefile", mask_shp_filepath,
                      mask_filepath])  # Convert .kml into .shp
        mask = fiona.open(mask_shp_filepath, "r")
        mask_srs = Proj(mask.crs["init"])

        mask_multi_polygon = []
        for feat in mask:
            mask_polygon = []
            for point in feat['geometry']['coordinates'][0]:
                long, lat = point[:2]  # one 2D point of the LinearRing
                x, y = transform(mask_srs, shp_srs, long, lat, always_xy=True)  # transform the point
                mask_polygon.append((x, y))
            mask_multi_polygon.append(mask_polygon)
            # mask_polygon is now in UTM proj, ready to compare to shapefile proj

        mask.close()

        return mask_multi_polygon

    def _read_annotations(self, annotations_filepath, raster, mask_filepath=None):
        shapefile = fiona.open(annotations_filepath, "r")
        shp_srs = Proj(shapefile.crs["init"])
        raster_srs = Proj(raster.crs)

        # --- Read and crop shapefile with mask if specified --- #
        if mask_filepath is not None:
            mask_multi_polygon = self._get_mask_multi_polygon(mask_filepath, shp_srs)
        else:
            mask_multi_polygon = None

        process_feat_partial = partial(process_feat, shp_srs=shp_srs, raster_srs=raster_srs,
                                       raster_transform=raster.transform, mask_multi_polygon=mask_multi_polygon)
        with Pool() as pool:
            out_polygons = list(
                tqdm(pool.imap(process_feat_partial, shapefile), desc="Process shp feature", total=len(shapefile),
                     leave=True))
        out_polygons = list(itertools.chain.from_iterable(out_polygons))

        shapefile.close()

        return out_polygons

    def process(self, metadata_list):
        progress_bar = tqdm(metadata_list, desc="Pre-process")
        for metadata in progress_bar:
            progress_bar.set_postfix(image=metadata["dirname"], status="Loading image")
            # Load image
            # image = skimage.io.imread(metadata["image_filepath"])
            raster = rasterio.open(metadata["image_filepath"])
            # print(raster)
            # print(dir(raster))
            # print(raster.meta)
            # exit()

            progress_bar.set_postfix(image=metadata["dirname"], status="Process shapefile")
            mask_filepath = metadata["mask_filepath"] if "mask_filepath" in metadata else None
            gt_polygons = self._read_annotations(metadata["gt_polygons_filepath"], raster, mask_filepath=mask_filepath)

            # Compute image mean and std
            b1, b2, b3 = raster.read()
            image = np.stack([b1, b2, b3], axis=-1)
            progress_bar.set_postfix(image=metadata["dirname"], status="Compute mean and std")
            image_float = image / 255
            mean = np.mean(image_float.reshape(-1, image_float.shape[-1]), axis=0)
            std = np.std(image_float.reshape(-1, image_float.shape[-1]), axis=0)

            if self.patch_size is not None:
                # Patch the tile
                progress_bar.set_postfix(image=metadata["dirname"], status="Patching")
                patch_stride = self.patch_stride if self.patch_stride is not None else self.patch_size
                patch_boundingboxes = image_utils.compute_patch_boundingboxes(image.shape[0:2],
                                                                              stride=patch_stride,
                                                                              patch_res=self.patch_size)
                for i, patch_boundingbox in enumerate(tqdm(patch_boundingboxes, desc="Process patches", leave=False)):
                    patch_gt_polygons = polygon_utils.crop_polygons_to_patch_if_touch(gt_polygons, patch_boundingbox)
                    if len(patch_gt_polygons) == 0:
                        # Do not save patches empty of polygons # TODO: keep empty patches?
                        break
                    patch_image = image[patch_boundingbox[0]:patch_boundingbox[2], patch_boundingbox[1]:patch_boundingbox[3], :]
                    sample = {
                        "dirname": metadata["dirname"],
                        "image": patch_image,
                        "image_mean": mean,
                        "image_std": std,
                        "gt_polygons": patch_gt_polygons,
                        "image_filepath": metadata["image_filepath"],
                    }

                    if self.pre_transform:
                        sample = self.pre_transform(sample)

                    filepath = os.path.join(self.processed_dir, sample["dirname"], "data.{:06d}.pt".format(i))
                    os.makedirs(os.path.dirname(filepath), exist_ok=True)
                    torch.save(sample, filepath)
            else:
                # Tile is saved as is
                sample = {
                    "dirname": metadata["dirname"],
                    "image": image,
                    "image_mean": mean,
                    "image_std": std,
                    "gt_polygons": gt_polygons,
                    "image_filepath": metadata["image_filepath"],
                }

                if self.pre_transform:
                    sample = self.pre_transform(sample)

                filepath = os.path.join(self.processed_dir, sample["dirname"], "data.{:06d}.pt".format(0))
                os.makedirs(os.path.dirname(filepath), exist_ok=True)
                torch.save(sample, filepath)

            flag_filepath = os.path.join(self.processed_dir, metadata["dirname"], "flag.json")
            flag = python_utils.load_json(flag_filepath)
            flag["done"] = True
            python_utils.save_json(flag_filepath, flag)

            raster.close()

    def _process(self):
        to_process_metadata_list = []
        for metadata in self.raw_sample_metadata_list:
            flag_filepath = os.path.join(self.processed_dir, metadata["dirname"], "flag.json")
            if os.path.exists(flag_filepath):
                flag = python_utils.load_json(flag_filepath)
                if not flag["done"]:
                    to_process_metadata_list.append(metadata)
            else:
                flag = {
                    "done": False
                }
                python_utils.save_json(flag_filepath, flag)
                to_process_metadata_list.append(metadata)

        if len(to_process_metadata_list) == 0:
            return

        print('Processing...')

        torch_lydorn.torch.utils.data.makedirs(self.processed_dir)
        self.process(to_process_metadata_list)

        path = os.path.join(self.processed_dir, 'pre_transform.pt')
        torch.save(torch_lydorn.torch.utils.data.__repr__(self.pre_transform), path)
        path = os.path.join(self.processed_dir, 'pre_filter.pt')
        torch.save(torch_lydorn.torch.utils.data.__repr__(self.pre_filter), path)

        print('Done!')

    def get(self, idx):
        filepath = self.processed_filepaths[idx]
        data = torch.load(filepath)
        data["patch_bbox"] = torch.tensor([0, 0, 0, 0])  # TODO: implement in pre-processing
        tile_name = os.path.basename(os.path.dirname(filepath))
        patch_name = os.path.basename(filepath)
        patch_name = patch_name[len("data."):-len(".pt")]
        data["name"] = tile_name + "." + patch_name
        return data


def process_feat(feat, shp_srs, raster_srs, raster_transform, mask_multi_polygon=None):
    out_polygons = []

    if feat['geometry']["type"] == "Polygon":
        poly = feat['geometry']['coordinates']
        polygons = process_polygon_feat(poly, shp_srs, raster_srs, raster_transform, mask_multi_polygon)
        out_polygons.extend(polygons)

    elif feat['geometry']["type"] == "MultiPolygon":
        for poly in feat['geometry']["coordinates"]:
            polygons = process_polygon_feat(poly, shp_srs, raster_srs, raster_transform, mask_multi_polygon)
            out_polygons.extend(polygons)

    return out_polygons


def process_polygon_feat(in_polygon, shp_srs, raster_srs, raster_transform, mask_multi_polygon=None):
    out_polygons = []
    points = in_polygon[0]  # TODO: handle holes
    # Intersect with mask_polygon if specified
    if mask_multi_polygon is not None:
        multi_polygon_simple = polygon_utils.intersect_polygons(points, mask_multi_polygon)
        if multi_polygon_simple is None:
            return out_polygons
    else:
        multi_polygon_simple = [points]

    for polygon_simple in multi_polygon_simple:
        new_poly = []
        for point in polygon_simple:
            x, y = point[:2]  # 740524.429227941 7175355.263524155
            x, y = transform(shp_srs, raster_srs, x, y)  # transform the point # 740520.728530676 7175320.732711278
            j, i = ~raster_transform * (x, y)
            new_poly.append((i, j))  # 2962.534577447921 2457.457061359659
        out_polygons.append(np.array(new_poly))

    return out_polygons


def get_seg_display(seg):
    seg_display = np.zeros([seg.shape[0], seg.shape[1], 4], dtype=np.float)
    if len(seg.shape) == 2:
        seg_display[..., 0] = seg
        seg_display[..., 3] = seg
    else:
        for i in range(seg.shape[-1]):
            seg_display[..., i] = seg[..., i]
        seg_display[..., 3] = np.clip(np.sum(seg, axis=-1), 0, 1)
    return seg_display


def main():
    # --- Params --- #
    config_name = "config.luxcarta_dataset"
    # --- --- #

    # Load config
    config = run_utils.load_config(config_name)
    if config is None:
        print_utils.print_error(
            "ERROR: cannot continue without a config file. Exiting now...")
        exit()

    # Find data_dir
    data_dir = python_utils.choose_first_existing_path(config["data_dir_candidates"])
    if data_dir is None:
        print_utils.print_error("ERROR: Data directory not found!")
        exit()
    else:
        print_utils.print_info("Using data from {}".format(data_dir))
    root_dir = os.path.join(data_dir, config["dataset_params"]["root_dirname"])

    # --- Transforms: --- #
    # --- pre-processing transform (done once then saved on disk):
    train_pre_transform = data_transforms.get_offline_transform(config,  # NOQA
                                                                augmentations=config["data_aug_params"]["enable"])
    eval_pre_transform = data_transforms.get_offline_transform(config, augmentations=False)  # NOQA
    # --- Online transform done on the host (CPU):
    train_online_cpu_transform = data_transforms.get_online_cpu_transform(config,
                                                                          augmentations=config["data_aug_params"][
                                                                              "enable"])
    eval_online_cpu_transform = data_transforms.get_online_cpu_transform(config, augmentations=False)  # NOQA
    # --- Online transform performed on the device (GPU):
    train_online_cuda_transform = data_transforms.get_online_cuda_transform(config,  # NOQA
                                                                            augmentations=config["data_aug_params"][
                                                                                "enable"])
    eval_online_cuda_transform = data_transforms.get_online_cuda_transform(config, augmentations=False)  # NOQA
    # --- --- #

    data_patch_size = config["dataset_params"]["data_patch_size"] if config["data_aug_params"]["enable"] else config["dataset_params"]["input_patch_size"]
    fold = "test"
    if fold == "train":
        dataset = LuxcartaBuildings(root_dir,
                                    transform=train_online_cpu_transform,
                                    patch_size=data_patch_size,
                                    patch_stride=config["dataset_params"]["input_patch_size"],
                                    pre_transform=data_transforms.get_offline_transform_patch(),
                                    fold="train",
                                    pool_size=config["num_workers"])
    elif fold == "test":
        dataset = LuxcartaBuildings(root_dir,
                                    transform=train_online_cpu_transform,
                                    pre_transform=data_transforms.get_offline_transform_patch(),
                                    fold="test",
                                    pool_size=config["num_workers"])

    print("# --- Sample 0 --- #")
    sample = dataset[0]
    print(sample["image"].shape)
    print(sample["gt_polygons_image"].shape)
    print("# --- Samples --- #")
    # for data in tqdm(dataset):
    #     pass

    data_loader = torch.utils.data.DataLoader(dataset, batch_size=10, shuffle=True, num_workers=0)
    print("# --- Batches --- #")
    for batch in tqdm(data_loader):
        print(batch["image"].shape)
        print(batch["gt_polygons_image"].shape)

        # Save output to visualize
        seg = np.array(batch["gt_polygons_image"][0]) / 255  # First batch
        seg = np.moveaxis(seg, 0, -1)
        seg_display = get_seg_display(seg)
        seg_display = (seg_display * 255).astype(np.uint8)
        skimage.io.imsave("gt_seg.png", seg_display)

        im = np.array(batch["image"][0])
        im = np.moveaxis(im, 0, -1)
        skimage.io.imsave('im.png', im)

        input("Enter to continue...")


if __name__ == '__main__':
    main()



"""
The GeoWATCH module

+----------------------------------------------------------+----------------------------------------------------------------+
| The GeoWATCH Gitlab Repo                                 | https://gitlab.kitware.com/computer-vision/geowatch/           |
+----------------------------------------------------------+----------------------------------------------------------------+
| The GeoWATCH Github Repo (Mirror)                        | https://github.com/Kitware/geowatch/                           |
+----------------------------------------------------------+----------------------------------------------------------------+
| Pypi                                                     | https://pypi.org/project/geowatch/                             |
+----------------------------------------------------------+----------------------------------------------------------------+
| Read the docs                                            | https://geowatch.readthedocs.io                                |
+----------------------------------------------------------+----------------------------------------------------------------+
| Slides                                                   | `Software Overview Slides`_  and `KHQ Demo Slides`_            |
+----------------------------------------------------------+----------------------------------------------------------------+

.. _Software Overview Slides: https://docs.google.com/presentation/d/125kMWZIwfS85lm7bvvCwGAlYZ2BevCfBLot7A72cDk8/

.. _KHQ Demo Slides: https://docs.google.com/presentation/d/1HKH_sGJX4wH60j8t4iDrZN8nH71jGX1vbCXFRIDVI7c/


Main modules of interest are:

    * :mod:`geowatch.cli`

    * :mod:`geowatch.mlops`

    * :mod:`geowatch.tasks`

Main Tasks:

    * :mod:`geowatch.tasks.fusion`

    * :mod:`geowatch.tasks.tracking`

Supported Feature Tasks:

    * :mod:`geowatch.tasks.cold`

    * :mod:`geowatch.tasks.depth`

    * :mod:`geowatch.tasks.depth_pcd`

    * :mod:`geowatch.tasks.dino_detector`

    * :mod:`geowatch.tasks.invariants`

    * :mod:`geowatch.tasks.landcover`

    * :mod:`geowatch.tasks.mae`

    * :mod:`geowatch.tasks.rutgers_material_seg_v2`

    * :mod:`geowatch.tasks.sam`

Also see:

    * :mod:`geowatch.gis`

    * :mod:`geowatch.geoannots`

    * :mod:`geowatch.demo`

    * :mod:`geowatch.stac`

    * :mod:`geowatch.utils`

    * :mod:`geowatch.utils.lightning_ext`

    * :mod:`geowatch.utils.lightning_ext.callbacks`

You probably wont need:

    * :mod:`geowatch.rc`

    * :mod:`geowatch.monkey`


.. code::

    # Useful environs

    # Phase 3
    DVC_DATA_DPATH=$(geowatch_dvc --tags='phase3_data' --hardware=auto)
    DVC_EXPT_DPATH=$(geowatch_dvc --tags='phase3_expt' --hardware=auto)
    echo "$DVC_DATA_DPATH"
    echo "$DVC_EXPT_DPATH"

    # Phase 2
    DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
    DVC_EXPT_DPATH=$(geowatch_dvc --tags='phase2_expt' --hardware=auto)
    HIGHRES_DVC_EXPT_DPATH=$(geowatch_dvc --tags='smart_drop7' --hardware=auto)
    DATA_DVC_DPATH=$DVC_DATA_DPATH
    EXPT_DVC_DPATH=$DVC_EXPT_DPATH

    # To get the above make sure you have run:
    geowatch_dvc add my_phase2_data_repo --path=<path-to-your-phase2-data-dvc-repo> --hardware=hdd --priority=100 --tags=phase2_data
    geowatch_dvc add my_phase2_expt_repo --path=<path-to-your-phase2-expt-dvc-repo> --hardware=hdd --priority=100 --tags=phase2_expt


.. code:: python

    # Useful Python Snippets
    import geowatch
    data_dvc_dpath = geowatch.find_dvc_dpath(tags='phase3_data', hardware='auto')
    expt_dvc_dpath = geowatch.find_dvc_dpath(tags='phase3_expt', hardware='auto')

"""
import os
import sys
import ubelt as ub
import warnings


__version__ = '0.20.0'


# ../dev/maintain/generate_authors.py
__author__ = 'GeoWATCH Developers, Kitware Inc., Jon Crall, David Joy, Matthew Bernstein, Connor Greenwell, Benjamin Brodie, Peri Akiva, Usman Rafique, Jacob DeRosa, Matthew Purri, Ajay Upadhyaya, Ji Won Suh, Jacob Birge, Ryan LaClair, Scott Workman, Dexter Lau, Sergii Skakun, Aram Ansary Ogholbake, Cohen Archbold, Bane Sullivan, Srikumar Sastry, Armin Hadzic'

__author_email__ = 'kitware@kitware.com, jon.crall@kitware.com'
__url__ = 'https://gitlab.kitware.com/computer-vision/geowatch'


os.environ['USE_PYGEOS'] = '0'


GEOWATCH_PREIMPORT_VARIANTS = {
    'variant1': ['geopandas', 'pyproj', 'gdal'],  # align-crs on horologic
    'variant2': ['pyproj', 'gdal'],               # CI machine
    'variant3': ['geopandas', 'pyproj'],          # delay gdal import
    'none': [],                                   # no pre-imports
    '0': [],                                      # no pre-imports
}


# Use this to debug import-time effects
_GEOWATCH_DEBUG = False


def _handle_hidden_commands():
    """
    Hidden developer features based on CLI flags or enviornment variables.
    The following table summarizes these options and their effect.

    +-------------------+------------------------+-----------------------------------------------+
    | CLI Flag          | Environment Variable   | Effect                                        |
    +-------------------+------------------------+-----------------------------------------------+
    | --warntb          | WARN_WITH_TRACEBACK    | Any warnings will emit a traceback to stderr  |
    +-------------------+------------------------+-----------------------------------------------+
    | --geowatch-debug  | _GEOWATCH_DEBUG        | Will print internal module setup info         |
    +-------------------+------------------------+-----------------------------------------------+

    """
    if ub.argflag('--geowatch-debug') or os.environ.get('_GEOWATCH_DEBUG', ''):
        global _GEOWATCH_DEBUG
        _GEOWATCH_DEBUG = 1
        if _GEOWATCH_DEBUG:
            print('Enabled geowatch debugging')

    if ub.argflag('--warntb') or os.environ.get('WARN_WITH_TRACEBACK', ''):
        import xdev
        xdev.make_warnings_print_tracebacks()

        if _GEOWATCH_DEBUG:
            print('Enable warning tracebacks')

    if _GEOWATCH_DEBUG:
        print('Finished handling hidden commands')


def _import_troublesome_module(modname):
    """
    Defines exactly how to import each troublesome (binary) module
    """
    if _GEOWATCH_DEBUG:
        print(f'Import troublesome module: {modname}')
    if modname == 'gdal':
        from osgeo import gdal as module
        gdal = module
        if not getattr(gdal, '_UserHasSpecifiedIfUsingExceptions', lambda: False)():
            if _GEOWATCH_DEBUG:
                print('Configuring GDAL to use exceptions')
            gdal.UseExceptions()
    elif modname == 'pyproj':
        import pyproj as module
        from pyproj import CRS  # NOQA
    elif modname == 'geopandas':
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', (
                '.*is incompatible with the GEOS version '
                'PyGEOS was compiled with.*'))
            import geopandas as module
    elif modname == 'rasterio':
        import rasterio as module
    elif modname == 'fiona':
        import fiona as module
    # elif modname == 'pygeos':
    #     import pygeos as module
    elif modname == 'torch':
        import torch as module
    elif modname == 'numpy':
        import numpy as module
    else:
        raise KeyError(modname)
    return module


def _is_running_a_fast_cli_tool():
    """
    Determine if we are running a fast cli tool.

    This is used to short circuit the pre-imports for certain command line
    tools so we don't incur the import time, which can be multiple seconds.

    TODO:
        Is there a better way to prevent certain entry points from executing
        the pre-imports?

    Notes:
        There is crazy sys.argv behavior with -m [SO42076706]_.

    References:
        .. [SO42076706] https://stackoverflow.com/questions/42076706/sys-argv-behavior-with-python-m

    Returns:
        bool: True if we should minimize startup time
    """
    if os.environ.get('_ARGCOMPLETE', ''):
        flag = True
    elif sys.argv and 'geowatch_dvc' in sys.argv[0] or 'geowatch' in sys.argv[0]:
        flag = True
    # elif sys.argv and len(sys.argv) == 1 and sys.argv[0] != '-c':
    #     # No args given case
    #     flag = True
    # elif sys.argv and sys.argv == ['-m']:
    #     flag = True
    elif sys.argv and ('--help' in sys.argv or '-h' in sys.argv):
        flag = True
    elif 'finish_install' in sys.argv:
        flag = True
    else:
        flag = False

    if _GEOWATCH_DEBUG:
        print(f'sys.argv={sys.argv}')
        print(f'Is running a fast CLI tool?: {flag}')

    return flag


def _execute_ordered_preimports():
    """
    The order in which certain modules with binary libraries are imported can
    impact runtime stability and can even cause crashes if a specific order
    isnt used.

    There are several known good configurations registered in the
    ``GEOWATCH_PREIMPORT_VARIANTS`` dictionary and the setting is
    controlled by the ``GEOWATCH_PREIMPORT`` environment variable.
    """
    if _GEOWATCH_DEBUG:
        print('Execute preordered imports')
    # Shorter alias because we are using it now
    GEOWATCH_PREIMPORT = os.environ.get('GEOWATCH_PREIMPORT', os.environ.get('WATCH_PREIMPORT', 'auto'))
    GEOWATCH_PREIMPORT = os.environ.get('GEOWATCH_HACK_IMPORT_ORDER', os.environ.get('WATCH_HACK_IMPORT_ORDER', GEOWATCH_PREIMPORT))

    if _GEOWATCH_DEBUG:
        print(f'GEOWATCH_PREIMPORT = {ub.urepr(GEOWATCH_PREIMPORT, nl=1)}')

    if not GEOWATCH_PREIMPORT:
        if _GEOWATCH_DEBUG:
            print('Fast return')
        return

    if GEOWATCH_PREIMPORT == 'auto':
        if RUNNING_FAST_CLI_TOOL:
            watch_preimport = None
        else:
            # This is the "known" best order for importing
            watch_preimport = GEOWATCH_PREIMPORT_VARIANTS['variant1']
    elif GEOWATCH_PREIMPORT in GEOWATCH_PREIMPORT_VARIANTS:
        watch_preimport = GEOWATCH_PREIMPORT_VARIANTS[GEOWATCH_PREIMPORT]
    elif GEOWATCH_PREIMPORT.lower() in {'0', 'false', 'no', ''}:
        watch_preimport = None
    else:
        watch_preimport = GEOWATCH_PREIMPORT.split(',')

    if watch_preimport is not None:
        for modname in watch_preimport:
            _import_troublesome_module(modname)


RUNNING_FAST_CLI_TOOL = _is_running_a_fast_cli_tool()

_handle_hidden_commands()
_execute_ordered_preimports()


# This can be removed if we bump the min scriptconfig version to 0.7.14
# from geowatch.monkey import monkey_scriptconfig   # NOQA
# monkey_scriptconfig.patch_0_7_14()


if 1:
    try:
        from geowatch.monkey import monkey_numpy  # NOQA
        # monkey_numpy.patch_numpy_dtypes()
        monkey_numpy.patch_numpy_2x()
    except Exception as ex:
        warnings.warn(f'issue with applying monkey patch: ex={ex}')


if 'hard-to-inspect-key' in vars():
    # Defined to hack jon's editor into autocompleting this.
    find_dvc_dpath = None


__devnotes__ = """

# Command to autogenerate lazy imports for this file
mkinit geowatch --lazy_loader --diff
mkinit geowatch --lazy_loader -w

# Debug import time
python -X importtime -c "import geowatch"
GEOWATCH_HACK_IMPORT_ORDER=variant3 python -X importtime -c "import geowatch"
GEOWATCH_HACK_IMPORT_ORDER=variant1 python -X importtime -c "import geowatch"
"""


# Choose which submodules (and which submodule attributes) to expose
__submodules__ = {
    '*': [],  # include all modules, but don't expose attributes
    'demo': ['coerce_kwcoco'],
    'utils': ['find_smart_dvc_dpath', 'find_dvc_dpath']
}


# --- AUTOGENERATED LAZY IMPORTS ---
# import lazy_loader  # NOQA


def lazy_import(module_name, submodules, submod_attrs, eager='auto'):
    import importlib
    import os
    name_to_submod = {
        func: mod for mod, funcs in submod_attrs.items()
        for func in funcs
    }

    def __getattr__(name):
        if name in submodules:
            attr = importlib.import_module(
                '{module_name}.{name}'.format(
                    module_name=module_name, name=name)
            )
        elif name in name_to_submod:
            submodname = name_to_submod[name]
            module = importlib.import_module(
                '{module_name}.{submodname}'.format(
                    module_name=module_name, submodname=submodname)
            )
            attr = getattr(module, name)
        else:
            raise AttributeError(
                'No {module_name} attribute {name}'.format(
                    module_name=module_name, name=name))
        globals()[name] = attr
        return attr

    eager_import_flag = False
    if eager == 'auto':
        eager_import_text = os.environ.get('EAGER_IMPORT', '')
        if eager_import_text:
            eager_import_text_ = eager_import_text.lower()
            if eager_import_text_ in {'true', '1', 'on', 'yes'}:
                eager_import_flag = True

        eager_import_module_text = os.environ.get('EAGER_IMPORT_MODULES', '')
        if eager_import_module_text:
            if __name__.lower() in eager_import_module_text.lower():
                eager_import_flag = True
    else:
        eager_import_flag = eager

    if eager_import_flag:
        for name in submodules:
            __getattr__(name)

        for attrs in submod_attrs.values():
            for attr in attrs:
                __getattr__(attr)
    return __getattr__


__getattr__ = lazy_import(
    __name__,
    submodules={
        'cli',
        'demo',
        'exceptions',
        'gis',
        'geoannots',
        'heuristics',
        'mlops',
        'monkey',
        'rc',
        'stac',
        'tasks',
        'utils',
    },
    submod_attrs={
        'demo': [
            'coerce_kwcoco',
        ],
        'utils': [
            'find_smart_dvc_dpath',
            'find_dvc_dpath',
        ],
    },
)


def __dir__():
    return __all__

__all__ = ['GEOWATCH_PREIMPORT_VARIANTS', 'cli', 'coerce_kwcoco', 'demo',
           'exceptions', 'find_dvc_dpath', 'find_smart_dvc_dpath', 'gis',
           'heuristics', 'mlops', 'monkey', 'rc', 'stac', 'tasks', 'utils']



"""
NOTE:
    THE IMPLEMENTATION HAS MOVED TO:
    https://gitlab.kitware.com/computer-vision/kwcoco_dataloader

    ON A DEV MACHINE IS MIGHT EXIST AT
    ~/code/kwcoco_dataloader/kwcoco_dataloader/tasks/fusion/datamodules/kwcoco_dataset.py
"""
from kwcoco_dataloader.tasks.fusion.datamodules.kwcoco_dataset import KWCocoVideoDatasetConfig  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.kwcoco_dataset import TruthMixin  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.kwcoco_dataset import GetItemMixin  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.kwcoco_dataset import IntrospectMixin  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.kwcoco_dataset import BalanceMixin  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.kwcoco_dataset import PreprocessMixin  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.kwcoco_dataset import MiscMixin  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.kwcoco_dataset import BackwardCompatMixin  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.kwcoco_dataset import KWCocoVideoDataset  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.kwcoco_dataset import worker_init_fn  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.kwcoco_dataset import _space_weights  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.kwcoco_dataset import sample_video_spacetime_targets  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.kwcoco_dataset import FailedSample  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.kwcoco_dataset import Modality  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.kwcoco_dataset import Domain  # NOQA
